"""
Generate maximum likelihood tree from an alignment given an evolutionary
model.
"""

import topiary

from ._raxml import run_raxml, RAXML_BINARY
from topiary.external._interface import prep_calc, write_run_information

import os, shutil, glob

def generate_ml_tree(previous_dir=None,
                     df=None,
                     model=None,
                     tree_file=None,
                     output=None,
                     overwrite=False,
                     threads=-1,
                     raxml_binary=RAXML_BINARY,
                     bootstrap=False):
    """
    Generate maximum likelihood tree from an alignment given an evolutionary
    model.

    Parameters
    ----------
    previous_dir : str, optional
        directory containing previous calculation. function will grab the the
        csv, model, and tree from the previous run. If this is not specified,
        `df`, `model`, and `tree_file` arguments must be specified.
    df : pandas.DataFrame or str, optional
        topiary data frame or csv written out from topiary df. Will override
        dataframe from `previous_dir` if specified.
    model : str, optional
        model (i.e. "LG+G8"). Will override model from `previous_dir`
        if specified.
    tree_file : str
        tree_file in newick format. Used as starting point for calculation.
        Will override tree from `previous_dir` if specified.
    output : str, optional
        output directory. If not specified, create an output directory
        with form "generate_ancestors_randomletters"
    overwrite : bool, default=False
        whether or not to overwrite existing output
    threads : int, default=-1
        number of threads to use. if -1, use all avaialable
    raxml_binary : str, optional
        what raxml binary to use
    bootstrap : bool, default=False
        whether or not to do bootstrap replicates

    Returns
    -------
    Python.core.display.Image or None
        if running in jupyter notebook, return Image of maximum likelihood tree;
        otherwise, return None
    """

    # Copy files in, write out alignment, move into working directory, etc.


    result = prep_calc(previous_dir=previous_dir,
                       df=df,
                       model=model,
                       tree_file=tree_file,
                       output=output,
                       overwrite=overwrite,
                       output_base="generate_ml_tree")

    df = result["df"]
    csv_file = result["csv_file"]
    model = result["model"]
    tree_file = result["tree_file"]
    alignment_file = result["alignment_file"]
    starting_dir = result["starting_dir"]

    other_args = []

    # If we're doing bootstrapping
    if bootstrap:
        algorithm = "--all"
        other_args.extend(["--bs-trees","autoMRE","--bs-write-msa"])
    else:
        algorithm = "--search"

    # Run raxml to create tree
    cmd = run_raxml(algorithm=algorithm,
                    alignment_file=alignment_file,
                    tree_file=tree_file,
                    model=model,
                    dir_name="working",
                    seed=True,
                    threads=threads,
                    raxml_binary=raxml_binary,
                    other_args=other_args)

    outdir = "output"
    os.mkdir(outdir)

    # Grab the final tree and store as tree.newick
    if bootstrap:
        shutil.copy(os.path.join("working","alignment.phy.raxml.support"),
                    os.path.join(outdir,"tree.newick"))
    else:
        shutil.copy(os.path.join("working","alignment.phy.raxml.bestTree"),
                    os.path.join(outdir,"tree.newick"))

    # Write run information
    write_run_information(outdir=outdir,
                          df=df,
                          calc_type="ml_tree",
                          model=model,
                          cmd=cmd)

    # Copy bootstrap results to the output directory
    if bootstrap:
        bs_out = os.path.join(outdir,"bootstrap_replicates")
        os.mkdir(bs_out)
        bsmsa = glob.glob(os.path.join("working","alignment.phy.raxml.bootstrapMSA.*.phy"))
        for b in bsmsa:
            number = int(b.split(".")[-2])
            shutil.copy(b,os.path.join(bs_out,f"bsmsa_{number:04d}.phy"))
        shutil.copy(os.path.join("working","alignment.phy.raxml.bootstraps"),
                    os.path.join(outdir,"bootstrap_replicates","bootstraps.newick"))

    print(f"\nWrote results to {os.path.abspath(outdir)}\n")

    # Leave working directory
    os.chdir(starting_dir)

    # Create plot holding tree
    ret = topiary.draw.ml_tree(run_dir=output,
                               output_file=os.path.join(output,
                                                        "output",
                                                        "summary-tree.pdf"))
    if topiary._in_notebook:
        return ret



"""
Draw a tree with ancestors colored by their relative posterior probability.
"""

import topiary
from topiary.external._interface import read_previous_run_dir

import os, re, copy

from ._base import load_trees, create_name_dict, setup_generic_tree_format, final_render
import ete3

def ancestor_tree(run_dir,
                  output_file=None,
                  tip_columns=["species","nickname"],
                  tip_name_separator="|",
                  fontsize=36,
                  circle_radius=0.025,
                  value_range=(0.6,1),R=(0.95,0),G=(0.95,0),B=(1,1),
                  width=800,space_between_taxa=10,line_width=4,
                  df=None):
    """
    Draw a tree with ancestors colored by their relative posterior probability.

    Parameters
    ----------
    run_dir : str
        directory containing an ancestral reconstruction run
    output_file : str, optional
        output file to write tree. If running in a notebook, will return to
        notebook and write to this file. If running in a notebook but not
        specified, do not write to a file. If not in a notebook and not
        specified, will write out ancestor-tree.pdf.
    tip_columns: list, default=["species","nickname"]
        label the tree tips as "|".join(tip_columns). For example, if
        tip_columns is ["species","nickname"], tips will have names like
        'Homo sapiens|LY96'.
    tip_name_separator : str, default="|"
        string to separate columns in tip names ("|" in tip_columns example
        above.)
    fontsize : float, default=36
        fontsize in points for labels
    circle_radius : float, default=0.025
        circle size for internal nodes (fraction of total width)
    value_range : tuple,None default=(0.6,1)
        tuple holding minimum and maximum values for support color map.
        (0,100) means miminum support value is 0, maximum support value
        is 100. If None, disable node color map.
    R : tuple, default=(0.95,0)
        tuple holding red channel values for lowest and highest supports.
    G : tuple, default=(0.95,0)
        tuple holding green channel values for lowest and highest supports.
    B : tuple, default=(1.0,1.0)
        tuple holding blue channel values for lowest and highest supports.
    width : int, default=800
        width of total tree in pixels
    space_between_taxa : int, default=10
        number of pixels between taxa, sets height.
    line_width : int, default=4
        width of lines used to draw tree (pixels)
    df : pandas.DataFrame or None, default=None
        topiary dataframe (overides whatever is in run_dir/output/dataframe.csv)

    Returns
    -------
    Python.core.display.Image or None
        if running in jupyter notebook, return Image; otherwise, return None
    """

    # Load data from previous run
    prev_run = read_previous_run_dir(run_dir)

    # Load the tree and relevant information from the ml_run_dir output
    # directory.
    T_label, T_pp = load_trees(prev_run_dir=run_dir,
                               tree_base_path=os.path.join("output","ancestors"),
                               tree_names=["ancestors_label.newick",
                                           "ancestors_pp.newick"],
                               tree_fmt=[1,1],
                               outgroup=prev_run["outgroup"])

    # If df not specified, get from the previous run
    if df is None:
        df = prev_run["df"]

    # Create dictionary mapping between uid and pretty name format
    name_dict = create_name_dict(df=df,
                                 tip_columns=tip_columns,
                                 separator=tip_name_separator)

    # Set up formats (color map for main nodes, tree format (ts) and generic
    # node format (ns)).
    cm, ts, ns = setup_generic_tree_format(num_leaves=len(T_label.get_leaves()),
                                           value_range=value_range,
                                           R=R,G=G,B=B,
                                           width=width,
                                           space_between_taxa=space_between_taxa,
                                           line_width=line_width)


    circle_radius = circle_radius*width

    # Grab the root from these trees. If the root as a posterior probability,
    # the annotation shifted when we rooted the tree. Grab the annotation and
    # put it on the inappropriately empty node that should have this annoation
    label_root = T_label.get_tree_root()
    pp_root = T_pp.get_tree_root()
    if pp_root.name != "":
        offset_root = (copy.deepcopy(label_root.name),copy.deepcopy(pp_root.name))
    else:
        offset_root = None

    # Traverse posterior probability and label trees simultaneously
    T_label_traversed = list(T_label.traverse())
    T_pp_traversed = list(T_pp.traverse())
    for i in range(len(T_label_traversed)):

        # Get T1 and T2 nodes (n and m)
        n = T_label_traversed[i]
        m = T_pp_traversed[i]

        # Set base style on node of T1
        n.set_style(ns)

        # Do not draw anything for root on ancestral reconstruction tree
        if n.is_root():
            continue

        # If this is not a leaf
        if not n.is_leaf():

            anc_name = n.name
            pp = m.name

            # If posterior probability is not defined for a node that is *not*
            # the root...
            if pp == "":

                # If the root was mislabeled above, stick the label on this node
                if offset_root is not None:
                    anc_name = offset_root[0]
                    pp = offset_root[1]
                    offset_root = None

            if cm is not None:

                try:
                    v = float(pp)
                    rgb = cm.hex(v)

                    # Draw circles
                    node_circle = ete3.CircleFace(radius=circle_radius,color=rgb,style="circle")
                    node_circle.margin_right=-circle_radius
                    n.add_face(node_circle,0,position="branch-right")

                except ValueError:
                    pass

            # Ancestor labels
            anc_label = ete3.TextFace(text=re.sub("anc","a",anc_name),fsize=fontsize)
            anc_label.inner_background.color = "white"
            anc_label.margin_right = 5
            anc_label.margin_top = 0
            n.add_face(anc_label,0,position="float")

        # If this is a leaf
        else:

            # Get the clean name (rather than uid)
            clean_name = name_dict[n.name]

            # Add text for clean name
            n.name = ""
            txt = ete3.TextFace(clean_name,fsize=fontsize)
            txt.margin_left = 4
            n.add_face(txt,0,position="branch-right")


    return final_render(T_label,ts,output_file,"ancestor-tree.pdf")



"""
Functions for reading and writing alignments to files.
"""

import topiary
from topiary import check, _private

import pandas as pd
import numpy as np

import os, re

def _validate_seq_writer(df,
                         out_file,
                         seq_column,
                         label_columns,
                         write_only_keepers,
                         empty_char,
                         clean_sequence,
                         overwrite):
    """
    Generic function for validating arguments used to validate sequences written
    out to files (such as fasta or phy).

    For argument parameter meanings, see write_fasta and write_phy.
    """

    df = topiary.check.check_topiary_dataframe(df)

    # Validate output file
    if type(out_file) is not str:
        err = f"\n\nout_file '{out_file}' should be a string.\n"
        raise ValueError(err)

    # Make sure seq column is sane
    try:
        if type(seq_column) is dict:
            raise TypeError
        df.loc[:,seq_column]
    except (KeyError,TypeError):
        err = f"seq_column '{seq_column}' not found\n."
        raise ValueError(err)

    # Make sure label columns are sane
    if hasattr(label_columns,"__iter__") and type(label_columns) is not type:
        if type(label_columns) is str:
            label_columns = [label_columns]
    else:
        err = "\nlabel_columns should be a list of column names in the\n"
        err += "dataframe.\n\n"
        raise ValueError(err)

    # Make sure the label columns are in the dataframe
    for c in label_columns:
        try:
            df[c]
        except (KeyError,TypeError):
            err = f"label_column '{c}' not recognized."
            err += "Should be a a column name in the dataframe.\n"
            raise ValueError(err)

    # Make sure label columns has uid in the first position
    try:
        label_columns.remove("uid")
    except ValueError:
        pass
    label_columns.insert(0,"uid")

    # Validate write_only_keepers argument
    try:
        bool(int(write_only_keepers))
    except (TypeError,ValueError):
        err = f"\nwrite_only_keepers '{write_only_keepers}' not recognized.\n"
        err += "Should be True or False.\n\n"
        raise ValueError(err)

    # Validate empty_char argument
    if empty_char is None:
        empty_char = []
    else:
        if type(empty_char) is not str:
            err = f"\nempty_char '{empty_char}' not recognized. This should be\n"
            err += "a string of characters\n\n"
            raise ValueError(err)

        empty_char = list(empty_char)

    # Validate clean_sequence argument
    try:
        bool(int(clean_sequence))
    except (TypeError,ValueError):
        err = f"\nclean_sequence '{clean_sequence}' not recognized. Should be\n"
        err += "True or False.\n\n"
        raise ValueError(err)

    # Validate overwrite argument
    try:
        bool(int(overwrite))
    except (TypeError,ValueError):
        err = f"\noverwrite '{overwrite}' not recognized. Should be\n"
        err += "True or False.\n\n"
        raise ValueError(err)

    # Check for file existance if overwrite = False
    if os.path.isfile(out_file):
        if not overwrite:
            err = f"\nout_file '{out_file}' already exists. To overwrite, set\n"
            err += "overwrite = True.\n\n"
            raise FileExistsError(err)

    return df, label_columns, empty_char


def write_fasta(df,out_file,seq_column="sequence",label_columns=["species","name"],
                write_only_keepers=True,empty_char="X-?",clean_sequence=False,
                overwrite=False):
    """
    Write a fasta file from a dataframe.

    Parameters
    ----------
    df: pandas.DataFrame
        data frame to write out
    out_file : str
        output file
    seq_column : str, default="sequence"
        column in data frame to use as sequence
    label_columns : list, default=["species","name"]
        list of columns to use for sequence labels
    write_only_keepers : bool, default=True
        whether or not to write only seq with keep == True
    empty_char : str or None
        string containing empty char. If the sequence is only empty char, do not
        write out. To disable check, set empty_char=None.
    clean_sequence : bool, default=False
        replace any non-aa characters with "-"
    overwrite : bool, default=False
        whether or not to overwrite an existing file

    Returns
    -------
    None
    """

    df, label_columns, empty_char = _validate_seq_writer(df,
                                                         out_file,
                                                         seq_column,
                                                         label_columns,
                                                         write_only_keepers,
                                                         empty_char,
                                                         clean_sequence,
                                                         overwrite)

    # Construct fasta output
    out = []
    for i in range(len(df)):
        row = df.iloc[i]

        if write_only_keepers:
            if not row.keep:
                continue

        # Create label for header
        h = []
        for c in label_columns:
            h.append(f"{row[c]}")
        h = "|".join(h)

        seq = row[seq_column]
        is_empty = len([s for s in seq if s not in empty_char]) == 0
        if seq == "" or seq is None or is_empty:
            continue

        # Replace non-aa characters with '-'
        if clean_sequence:
            seq = re.sub("[^ACDEFGHIKLMNPQRSTVWY-]","-",seq)

        out.append(f">{h}\n{seq}\n")

    if len(out) == 0:
        err = "\nNo sequences written out. Check your settings and the dataframe.\n\n"
        raise RuntimeError(err)

    # Write output
    f = open(out_file,"w")
    f.write("".join(out))
    f.close()

def write_phy(df,
              out_file,
              seq_column="alignment",
              write_only_keepers=True,
              empty_char="X-?",
              clean_sequence=False,
              overwrite=False):
    """
    Write a .phy file from a dataframe. Uses the uid as the sequence name. All
    sequences must have the same length.

    Parameters
    ----------
    df: pandas.DataFrame
        data frame to write out
    out_file : str
        output file
    seq_column : str, default="alignment"
        column in data frame to use as sequence
    label_columns : list, default=["species","name"]
        list of columns to use for sequence labels
    write_only_keepers : bool, default=True
        whether or not to write only seq with keep == True
    empty_char : str or None
        string containing empty char. If the sequence is only empty char, do not
        write out. To disable check, set empty_char=None.
    clean_sequence : bool, default=False
        replace any non-aa characters with "-"
    overwrite : bool, default=False
        whether or not to overwrite an existing file

    Returns
    -------
    None
    """

    label_columns = ["uid"]

    df, label_columns, empty_char = _validate_seq_writer(df,
                                                         out_file,
                                                         seq_column,
                                                         label_columns,
                                                         write_only_keepers,
                                                         empty_char,
                                                         clean_sequence,
                                                         overwrite)

    if write_only_keepers:
        num_to_write = np.sum(df.keep)
    else:
        num_to_write = len(df.keep)

    all_lengths = []
    for i in range(len(df)):
        try:
            # If this is not being kept, don't record
            if write_only_keepers and np.logical_not(df.keep.iloc[i]):
                raise TypeError

            # If sequence is length 0 (or some other wacky non-len compatible
            # input) don't record.
            l = len(df[seq_column].iloc[i])
            if l == 0:
                raise TypeError

            # Record length
            all_lengths.append(l)
        except TypeError:
            if not df.keep.iloc[i] and write_only_keepers:
                pass
            else:
                err = "\nRow does not have sequence\n\n"
                err += f"...{df.iloc[i]}\n"
                raise ValueError(err)

    all_lengths = set(all_lengths)
    if len(all_lengths) == 0:
        err = "\nno sequences found\n\n"
        raise ValueError(err)

    if len(all_lengths) != 1:
        err = "\nnot all rows have the same sequence length\n\n"
        raise ValueError(err)

    # Finally, get length of alignment
    ali_length = list(all_lengths)[0]

    # Construct phy output
    out = []
    for i in range(len(df)):
        row = df.iloc[i]

        if write_only_keepers:
            if not row.keep:
                continue

        h = row["uid"]
        seq = row[seq_column]
        is_empty = len([s for s in seq if s not in empty_char]) == 0
        if seq == "" or seq is None or is_empty:
            num_to_write -= 1
            continue

        # Replace non-aa characters with '-'
        if clean_sequence:
            seq = re.sub("[^ACDEFGHIKLMNPQRSTVWY-]","-",seq)

        out.append(f"{h}\n{seq}\n")

    if len(out) == 0:
        err = "\nNo sequences written out. Check your settings and the dataframe.\n\n"
        raise RuntimeError(err)


    out.insert(0,f"{num_to_write}  {ali_length}\n\n")

    # Write output
    f = open(out_file,"w")
    f.write("".join(out))
    f.close()

def read_fasta_into(df,fasta_file,load_into_column="alignment",unkeep_missing=True):
    """
    Load sequences from a fasta file into an existing topiary dataframe. This
    function expects the fasta file to have names formated like >uid|other stuff.
    It will match the uid in the fasta file with the uid in the topiary dataframe.
    If a uid is not in the dataframe, the function will raise an error. 

    Parameters
    ----------
    df : pandas.DataFrame
        topiary data frame
    fasta_file : str
        a fasta file with headers formatted like >uid|other stuff
    load_into_column : str, default="alignment"
        what column in the dataframe to load the sequences into
    unkeep_missing : bool, default=True
        set any sequences in the dataframe tht are not in the fasta file to
        keep=False. This allows the user to delete sequences from the alignment
        and have that reflected in the dataframe.

    Return
    ------
    pandas.DataFrame
        topiary dataframe with sequences now in load_into_column
    """

    # Create data frame and make sure it has the column in which to load
    new_df = check.check_topiary_dataframe(df)
    try:
        new_df[load_into_column]
    except KeyError:
        new_df[load_into_column] = None

    # Go through the fasta file and get sequences
    headers = []
    uids = []
    seqs = []
    with open(fasta_file) as f:
        for line in f:
            if line.startswith(">"):
                headers.append(line.strip())
                uids.append(line[1:].split("|")[0].strip())
                seqs.append([])
            else:
                seqs[-1].append(line.strip())

    if len(uids) != len(set(uids)):
        err = "Not all uids unique in this fasta file\n"
        raise ValueError(err)

    final_seqs = []
    for s in seqs:
        final_seqs.append("".join(s))
    uid_to_index = dict(zip(df.uid,df.index))

    # Load sequences from fasta into data frame
    loaded_seq = {}
    for i in range(len(seqs)):

        try:
            index = uid_to_index[uids[i]]
            new_df.loc[index,load_into_column] = final_seqs[i]
            loaded_seq[index] = None
        except KeyError:
            err = f"Could not map the sequence titled {headers[i]} to an index\n"
            err += "in the data frame. This function expects the sequence titles\n"
            err += "in a fasta file to have the format:\n"
            err += ">uid|other stuff here\n"
            err += f"The parsed uid ({uids[i]}) is not in the dataframe!\n\n"
            raise ValueError(err)

    # If requested, set all sequences not in alignment to Keep = False
    if unkeep_missing:
        for i in list(new_df.index):
            try:
                loaded_seq[i]
            except KeyError:
                new_df.loc[i,"keep"] = False
                new_df.loc[i,load_into_column] = pd.NA

    # Return dataframe with final sanity check to make sure uid stayed unique
    return check.check_topiary_dataframe(new_df)



"""
Functions for reading and writing dataframes.
"""

import topiary
from topiary import check

import pandas as pd
import numpy as np

import os

def read_dataframe(input,remove_extra_index=True):
    """
    Read a topiary spreadsheet. Handles .csv, .tsv, .xlsx/.xls. If extension is
    not one of these, attempts to parse text as a spreadsheet using
    `pandas.read_csv(sep=None)`.

    Parameters
    ----------
    input : pandas.DataFrame or str
        either a pandas dataframe OR the filename to read in.
    remove_extra_index : bool, default=True
        look for the 'Unnamed: 0' column that pandas writes out for
        pandas.to_csv(index=True) and, if found, drop column.

    Returns
    -------
    pandas.DataFrame
        validated topiary dataframe
    """

    # If this is a string, try to load it as a file
    if type(input) is str:

        filename = input

        ext = filename.split(".")[-1].strip().lower()

        if ext in ["xlsx","xls"]:
            df = pd.read_excel(filename)
        elif ext == "csv":
            df = pd.read_csv(filename,sep=",")
        elif ext == "tsv":
            df = pd.read_csv(filename,sep="\t")
        else:
            # Fall back -- try to guess delimiter
            df = pd.read_csv(filename,sep=None,engine="python")

    # If this is a pandas dataframe, work in a copy of it.
    elif type(input) is pd.DataFrame:
        df = input.copy()

    # Otherwise, fail
    else:
        err = f"\n\n'input' {input} not recognized. Should be the filename of\n"
        err += "spreadsheet or a pandas dataframe.\n"
        raise ValueError(err)

    # Look for extra index column that pandas writes out (in case user wrote out
    # pandas frame manually, then re-read). Looks for first column that is
    # Unnamed and has values [0,1,2,...,L]
    if remove_extra_index:
        if df.columns[0].startswith("Unnamed:"):
            possible_index = df.loc[:,df.columns[0]]
            if np.issubdtype(possible_index.dtypes,int):
                if np.array_equal(possible_index,np.arange(len(possible_index),dtype=int)):
                    df = df.drop(columns=[df.columns[0]])

    # Validate the dataframe
    df = check.check_topiary_dataframe(df)

    return df

def write_dataframe(df,out_file,overwrite=False):
    """
    Write a dataframe to an output file. The type of file written depends on the
    extension of out_file. If .csv, write comma-separated. If .tsv, write tab-
    separated. If .xlsx, write excel. Otherwise, write as a .csv file.

    Parameters
    ----------
    df : pandas.DataFrame
        topiary dataframe
    out_file : str
        output file name
    overwrite : bool, default=False
        whether or not to overwrite an existing file

    Returns
    -------
    None
    """

    # Validate dataframe before writing out
    df = topiary.check.check_topiary_dataframe(df)

    # Validate output file
    if type(out_file) is not str:
        err = f"\n\nout_file '{out_file}' should be a string.\n"
        raise ValueError(err)

    ext = out_file.split(".")[-1]
    if ext not in ["csv","tsv","xlsx"]:
        print("\n\nOutput file extension not recognized. Will write as csv.\n\n")
        ext = "csv"

    if os.path.exists(out_file):
        if not overwrite:
            err = f"\nout_file '{out_file}' exists. To overwrite, set overwrite to True\n\n"
            raise FileExistsError(err)
        else:
            os.remove(out_file)

    # Write out appropriate file type
    if ext == "csv":
        df.to_csv(out_file,sep=",",index=False)
    elif ext == "tsv":
        df.to_csv(out_file,sep="\t",index=False)
    else:
        df.to_excel(out_file,index=False)



"""
Sample sequences from a topiary dataframe in a taxonomically informed way.
"""

import topiary
from topiary.quality import score_alignment
from topiary import check

import ete3

import numpy as np

import copy

def _prep_species_tree(df,paralog_column):
    """
    Create a species tree with paralogs loaded into paralogs attribute.

    Parameters
    ----------
        df: topiary dataframe
        paralog_column: column in dataframe containing paralog calls for each
                        sequence.

    Return
    ------
        annotated species tree (ete3)
    """

    if paralog_column not in df.columns:
        err = f"\nparalog_column '{paralog_column}' not in dataframe.\n"
        raise ValueError(err)

    # Make sure df has ott
    if "ott" not in df.columns:
        df = topiary.opentree.get_ott_id(df)

    # Get rid of almost identical sequences within each species
    species_tree = topiary.opentree.get_species_tree(df)

    all_paralogs = list(np.unique(df.loc[:,paralog_column]))
    paralogs_seen = dict([(p,[]) for p in all_paralogs])
    for leaf in species_tree.get_leaves():
        leaf.paralogs = copy.deepcopy(paralogs_seen)

        this_df = df.loc[df.loc[:,"uid"].isin(leaf.uid),:]
        for i in range(len(this_df)):
            idx = this_df.index[i]
            leaf.paralogs[this_df.loc[idx,paralog_column]].append(this_df.loc[idx,"uid"])

    return species_tree

def _divide_budget_among_paralogs(T,overall_budget):
    """
    Given some overall number of sequences desired, figure out how many to
    assign to each paralog. Decides how to do this in a tree-weighted
    fashion.

    Parameters
    ----------
        T: ete3 tree with leaves that have the `paralogs` dictionary holding
           lists of uid with paralog names as keys. Assumes all branch
           lengths are identical.
        overall_budget: integer number of sequences to keep adding up all
                        all paralogs.

    Return
    ------
        paralog_budget dictionary holding total number of sequences to keep for
        each paralog
    """

    # Get all unique paralogs in the tree
    paralogs = []
    for leaf in T.get_leaves():
        paralogs.extend(list(leaf.paralogs.keys()))
    paralogs = list(set(paralogs))

    # Figure out a weighted total for each paralog. The weights have two
    # components. Number of paralogs seen (higher weight) and the distance
    # from the ancestor (lower weight). If an organism has two called copies
    # of some paralog and is four splits from the tree ancestor, it will
    # recieve a weight of 2*(0.5^4) = 0.125. If an organism has one called
    # copy of the same paralog, but directly splits off the ancestor, it
    # will have a weight of 1*(0.5^1) = 0.5. This strategy is to ensure we
    # have an adequte number of seqeunces to capture both early duplications
    # and late expansions.
    paralog_totals = dict([(p,0) for p in paralogs])
    anc = T.get_common_ancestor(T.get_leaves())
    for leaf in T.get_leaves():
        distance = leaf.get_distance(anc)
        weight = np.power(0.5,distance)

        for p in paralogs:
            num_p = len(leaf.paralogs[p])
            paralog_totals[p] += num_p*weight

    # Given our overall budget and the weights on each paralog seen, give
    # each paralog a total budget (paralog_weight/sum(paralog_weights) * overall).
    paralog_budget = {}
    all_total = sum([paralog_totals[p] for p in paralog_totals])
    for p in paralog_totals:
        fx = paralog_totals[p]/all_total
        paralog_budget[p] = int(np.round(fx*overall_budget,0))

    return paralog_budget


def _get_sequence_budgets(T,total_budget):
    """
    Figure out the sequence budget for each node in a tree. Starts at the
    ancestor of the tree and then splits sequences as evenly as possible
    among descendants. Assigns a 'budget' attribute to each node indicating
    how many sequences are budgeted for that node and its descendants.

    Parameters
    ----------
        T: ete3 with `sequences` list associated with each node.
        total_budget: total number of sequences to budget over the whole tree

    Return
    ------
        T updated with .budget and .num_seq attributes
    """

    # Starts at ancestor and moves up
    for current_node in T.traverse(strategy="levelorder"):

        # Get ancestor of clade containing sisters to current group
        sisters = current_node.get_sisters()
        num_sister_clades = len(sisters)
        if num_sister_clades == 1:
            sister_node = sisters[0]
        else:
            # This is the root. Record the total budget available and continue
            if num_sister_clades == 0:
                current_node.budget = total_budget
                current_node.num_seq = sum([len(l.sequences) for l in current_node.get_leaves()])
                continue
            else:
                err = "\nInput tree must be rooted and fully resolved (no polytomies)\n\n"
                raise ValueError(err)

        # If we've already visited this split, don't do anything
        if hasattr(current_node,"budget") and hasattr(sister_node,"budget"):
            continue

        # Get available budget from parent of current split
        parent_node = T.get_common_ancestor(current_node,sister_node)
        avail_budget = parent_node.budget

        # Get the number of sequences down the current and sister splits
        current_seq = sum([len(l.sequences) for l in current_node.get_leaves()])
        current_node.num_seq = current_seq

        sister_seq = sum([len(l.sequences) for l in sister_node.get_leaves()])
        sister_node.num_seq = sister_seq

        # Only one sequence in the budget. Only assign if there is only one sequence
        # left. Otherwise, set to 0.
        if avail_budget == 1:
            if current_seq == 1 and sister_seq == 0:
                current_node.budget = 1
                sister_node.budget = 0
            elif current_seq == 0 and sister_seq == 1:
                current_node.budget = 0
                sister_node.budget = 1
            else:
                current_node.budget = 0
                sister_node.budget = 0

        # Try to partition the budget
        else:

            # We have enough budget to accomodate all sequences.
            if current_seq + sister_seq <= avail_budget:

                current_budget = current_seq
                sister_budget = sister_seq

            # We do not have the budget for all sequences.
            else:

                # Amount each would get in an even split
                even_split = avail_budget//2
                remainder = avail_budget % 2

                # If there are fewer current_seq than what they would get with an even
                # split, keep all of them and assign remaining budget to sister.
                if current_seq < even_split:
                    current_budget = current_seq
                    sister_budget = avail_budget - current_budget

                # If there are fewer sister_seq than what they would get with an even
                # split, keep all of them and assign remaining budget to current.
                elif sister_seq < even_split:
                    sister_budget = sister_seq
                    current_budget = avail_budget - sister_budget

                # Both are above what they would get from an even split. Give them both
                # a budget of even_split, with any remainder going to the node with
                # more descendant sequences.
                else:
                    sister_budget = even_split
                    current_budget = even_split

                    if current_seq >= sister_seq:
                        current_budget += remainder
                    else:
                        sister_budget += remainder

            # Assign the allocated budgets to the node
            current_node.budget = current_budget
            sister_node.budget = sister_budget

    return T

def _identify_merge_blocks(T):

    T = T.copy()
    merge_blocks = []
    for counter, leaf in enumerate(T.get_leaves()):

        # Already placed into a merge block
        if leaf.budget == -1:
            continue

        # If we have enough budget to take every sequence
        if leaf.num_seq <= leaf.budget:

            # No sequences -- don't do anything
            if leaf.num_seq == 0:
                continue

            # Record that we want to merge all sequences on this leaf
            merge_uid = list(leaf.sequences)
            merge_blocks.append((len(merge_uid),merge_uid,leaf))
            leaf.budget = -1

        else:

            # If the leaf has some budget, we're taking a subset of the
            # sequences on this leaf.
            if leaf.budget > 0:
                merge_uid = list(leaf.sequences)
                merge_blocks.append((leaf.budget,merge_uid,leaf))
                leaf.budget = -1

            # If the budget is zero, we need to go down the tree to find
            # a group of leaves to merge
            else:
                for anc in leaf.get_ancestors():

                    # If we have a non-zero budget, merge all descendants
                    if anc.budget > 0:

                        budget = anc.budget
                        merge_uid = []
                        for m in anc.get_leaves():
                            if m.budget == -1:
                                err = "We hit a descendant leaf of that\n"
                                err += "has already been placed into a merge. This\n"
                                err += "should not be possible for a monophyletic\n"
                                err += "tree!\n\n"
                                raise RuntimeError(err)
                            merge_uid.extend(m.sequences)
                            m.budget = -1

                        merge_blocks.append((budget,merge_uid,anc))

                        # Break out of get_ancestors loop
                        break

    # Sanity check on merge
    for l in T.get_leaves():
        if l.num_seq > 0:
            assert l.budget == -1

    return merge_blocks

def taxonomic_sample(df,
                     paralog_column="recip_paralog",
                     target_seq_number=500,
                     even_paralog_split=True,
                     key_species=[],
                     within_species_redundancy_cutoff=0.99,
                     sparse_column_cutoff=0.95,
                     sparse_run_length_keep_percentile=0.99,
                     fx_missing_dense_cutoff=0.75,
                     align_trim=(0.05,0.95),
                     verbose=False):
    """
    Sample sequences from a topiary dataframe in a taxonomically informed way.

    Parameters
    ----------
    df : pandas.DataFrame
        topiary dataframe
    paralog_column : str, default="recip_paralog"
        column holding preliminary paralog calls.
    target_seq_number : int, default=500
        number of sequences to aim for in the final dataset
    even_paralog_split : bool, default=True
        try to keep an even number of paralogs. If False, select the relative
        number of each paralog to keep based on its frequency in the dataframe
    key_species : list, default=[]
        list of species (binomial names) that will not have sequences removed
        (unless there are duplicate sequences for that species with identity >
        within_species_redundancy_cutoff).
    within_species_redundancy_cutoff : float, default=0.99
        merge sequences with sequence identity above cutoff when removing
        redundancy of sequences within species.
    sparse_column_cutoff : float, default=0.95
        when checking alignment quality, a column is sparse if it has gaps in
        more than sparse_column_cutoff sequences.
    sparse_run_length_keep_percentile : float, 0.99
        when checking alignment quality, remove the sequences with the longest
        insertions. Toss the longest (1 - sparse_run_length_keep_percentile)*num_sequences
        sequences.
    fx_missing_dense_cutoff : float, default=0.75
        when checking alignment quality, remove sequences that are missing more
        than 1 - fx_missing_dense_cutoff of the dense (that is, not sparse,
        columns).
    align_trim : tuple, default=(0.05,0.95)
        when checking alignment quality, do not score the first and last parts
        of the alignment. Interpreted like a slice, but with percentages.
        (0.0,1.0) would not trim; (0.05,0,98) would trim the first 0.05 off the
        front and the last 0.02 off the back.
    verbose : bool, default=False
        output verbosity
    """

    df = check.check_topiary_dataframe(df)

    try:
        df.loc[:,paralog_column]
    except KeyError:
        err = f"\nparalog_column '{paralog_column}' not found in dataframe.\n\n"
        raise ValueError(err)

    target_seq_number = check.check_int(target_seq_number,
                                                 "target_seq_number",
                                                 minimum_allowed=1)

    key_species = check.check_iter(key_species,
                                               "key_species")

    within_species_redundancy_cutoff = check.check_float(within_species_redundancy_cutoff,
                                                         "within_species_redundancy_cutoff",
                                                         minimum_allowed=0.0,
                                                         maximum_allowed=1.0)

    sparse_run_length_keep_percentile = check.check_float(sparse_run_length_keep_percentile,
                                                          "sparse_run_length_keep_percentile",
                                                          minimum_allowed=0.0,
                                                          maximum_allowed=1.0)

    fx_missing_dense_cutoff = check.check_float(fx_missing_dense_cutoff,
                                                "fx_missing_dense_cutoff",
                                                minimum_allowed=0.0,
                                                maximum_allowed=1.0)

    fx_missing_dense_cutoff = 1 - fx_missing_dense_cutoff

    ## align_trim passed directly to score_alignment and validated there

    verbose = check.check_bool(verbose,"verbose")

    # Drop unkept columns
    df = df.loc[df.keep,:]

    # How many sequences we start with
    starting_keep = np.sum(df.keep)
    print(f"Starting with {starting_keep} sequences.\n",flush=True)

    if within_species_redundancy_cutoff < 1.0:
        print("Removing nearly identical sequences within species.\n",flush=True)
        df = topiary.quality.remove_redundancy(df,
                                               cutoff=within_species_redundancy_cutoff,
                                               only_in_species=True)

        if np.sum(df.keep) == 0:
            err = "redundnacy pass removed all sequences!\n"
            raise ValueError(err)

    # Drop unkept columns
    df = df.loc[df.keep,:]

    # Update dataframe with alignment score information
    df["sparse_run_length"] = 0.0
    df["fx_missing_dense"] = 0.0

    # Iterate over all paralogs.
    paralogs = set(df.loc[:,paralog_column])
    for p in paralogs:

        # Create a dataframe with this paralog only
        paralog_mask = df.loc[:,paralog_column] == p
        this_df = df.loc[paralog_mask,:]

        # Align the sequenes within this paralog
        print(f"Performing initial alignment of paralog {p}.\n",flush=True)
        this_df = topiary.run_muscle(this_df,super5=True)

        print(f"\nRemoving poorly aligned sequences.\n",flush=True)

        # Score alignment
        this_df = score_alignment(this_df,
                                  alignment_column="alignment",
                                  align_trim=align_trim,
                                  sparse_column_cutoff=sparse_column_cutoff)

        df.loc[paralog_mask,"sparse_run_length"] = this_df["sparse_run_length"]
        df.loc[paralog_mask,"fx_missing_dense"] = this_df["fx_missing_dense"]

        # Remove hardest-to-align to align sequences from consideration
        tmp = np.array(this_df.sparse_run_length)
        slicer = int(np.round(len(tmp)*sparse_run_length_keep_percentile,0))
        sparse_run_length_cutoff = tmp[np.argsort(tmp)][slicer]

        keep_mask = np.product((this_df.sparse_run_length < sparse_run_length_cutoff,
                                this_df.fx_missing_dense < fx_missing_dense_cutoff),axis=0)

        #df.loc[paralog_mask,"keep"] = np.logical_and(df.loc[paralog_mask,"keep"],
        #                                             keep_mask)

        if np.sum(keep_mask) == 0:
            err = f"alignment quality pass removed all sequences for paralog {p}!\n"
            raise ValueError(err)


    print(f"\nInitial sequence quality control: {starting_keep} --> {np.sum(df.keep)} sequences.\n",flush=True)

    # Construct species tree annotated with paralogs at tips
    species_tree = _prep_species_tree(df,paralog_column=paralog_column)

    # Get the sequence budget for all paralogs
    if even_paralog_split:
        paralog_totals = {}
        for p in paralogs:
            paralog_totals[p] = np.sum(df.loc[:,paralog_column] == p)
        all_total = sum([paralog_totals[p] for p in paralog_totals])

        paralog_budget = {}
        for p in paralog_totals:
            fx = paralog_totals[p]/all_total
            paralog_budget[p] = int(np.round(fx*target_seq_number,0))
    else:
        paralog_budget = _divide_budget_among_paralogs(species_tree,target_seq_number)

    print("Approximate number of paralogs to keep")
    for p in paralog_budget:
        print(f"    {p}: {paralog_budget[p]}")
    print("",flush=True)

    # Go through each paralog
    uid_to_keep = []
    for p in paralog_budget:

        # Create a tree that has only paralog p in it
        T = species_tree.copy()
        for leaf in T.get_leaves():
            leaf.sequences = tuple(leaf.paralogs[p])
            leaf.name = f"{leaf.name} ({len(leaf.sequences)})"

        # Figure out how to distribute the total budget across the tree
        _get_sequence_budgets(T,paralog_budget[p])

        # Get list of blocks of sequences to merge
        merge_blocks = _identify_merge_blocks(T)

        # Merge each merge block
        for m in merge_blocks:
            budget = m[0]
            uid = m[1]

            this_mask = df.loc[:,"uid"].isin(uid)
            this_df = df.loc[this_mask,:]

            # If a key_species is in the block, take it.
            in_species_mask = this_df.species.isin(key_species)
            if np.sum(in_species_mask) >= budget:
                this_uid = list(this_df.loc[in_species_mask,"uid"])
            else:

                # Sort by best aligner
                a = np.array(this_df.fx_missing_dense/np.sum(this_df.fx_missing_dense))
                b = np.array(this_df.sparse_run_length/np.sum(this_df.sparse_run_length))

                sort_order = np.argsort(a + b)
                this_uid = np.array(this_df.uid.iloc[sort_order])[:budget]

            if verbose:

                species = this_df.loc[this_df.uid.isin(this_uid),"species"]

                print("---------------------------------------------------------")
                if budget == len(uid):
                    print("KEEPING")
                else:
                    print("MERGING")
                print("---------------------------------------------------------")

                print(m[2])
                print()
                print(f"{p}",",".join(species))
                print(flush=True)

            uid_to_keep.extend(this_uid)


    # Update dataframe keep with merge results
    keep_mask = df.loc[:,"uid"].isin(uid_to_keep)
    df.loc[keep_mask,"keep"] = True
    df.loc[np.logical_not(keep_mask),"keep"] = False
    if "always_keep" in df:
        df.loc[df.always_keep,"keep"] = True

    print("Final number of sequences:")
    print(sum(df.keep))
    print()

    return df



"""
Function to check/process topiary dataframes.
"""

from topiary import _private

import pandas as pd
import numpy as np

import re

def check_topiary_dataframe(df):
    """
    Check to make sure topiary dataframe is valid, adding standard columns if
    necessary. Edits and returns a copy of the input dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        topiary dataframe to validate

    Returns
    -------
    pandas.DataFrame
        validated copy of input dataframe

    Notes
    -----
    + Checks for required columns 'species', 'name', and 'sequence'. Casts these
      entries as strings and does not allow them to be None/NaN/pandas.NA.
    + Deletes duplicated rows.
    + Deletes empty rows.

    + `keep`:
        + Must be boolean.
        + If present but not boolean, try to coerce it into boolean.
            + if str, map regex [1yt] --> True and regex[0nf] --> False
            + if int, map !0 --> True and 0 --> False
            + if float, map !~0 --> True and ~0 --> False
        + If keep is not present, add and set all to True.
    + `uid`:
        + Must be unique str of length 10 consisting only of [a-z][A-Z]
        + If present:
            + If missing values, fill them in
            + If values do not meet rules, replace them with uid that do
            + If values are not unique, update them to be unique
        + If not present, create them.
    + `ott`: If present, makes sure it has the format ottINTEGER.
    + `alignment`: if present, makes sure that all columns are the same width,
      and removes any gaps-only columns.

    + Enforces following column order:
        + `nickname` (if present)
        + `keep`
        + `species`
        + `name`
        + `sequence`
        + all other columns, in order they came in
    """

    # Make sure type is right
    if type(df) is not pd.DataFrame:
        err = "\ndf must be a pandas dataframe.\n\n"
        raise ValueError(err)

    # Drop duplicates
    df = df.drop_duplicates(ignore_index=True)

    # Get mask for cells that are either null or nan
    bad_mask = np.logical_or(pd.isnull(df),pd.isna(df))
    idx = list(df.index)
    final_mask = []
    for i in range(len(bad_mask)):

        # Grab row
        row = df.loc[idx[i],:]

        # Get all non-null and non-na entries from the row
        mask = np.logical_not(np.logical_or(pd.isnull(row),pd.isna(row)))
        not_null = row[mask]

        # Default to not keeping a row
        keep = False

        # If there are any non-null entries left...
        if len(not_null) > 0:

            # Loop will end up with keep = False if only columns remaining are
            # type string and "".
            for n in not_null:

                # If an empty string, this is basically a null.
                if type(n) is str and n.strip() == "":
                    continue

                # If we get here, we have some kind of non-empty cell. Break out
                # and keep row
                keep = True
                break

        final_mask.append(keep)

    final_mask = np.array(final_mask,dtype=bool)

    # Let user know we're dropping lines
    if np.sum(np.logical_not(final_mask)) > 0:
        lines = ",".join(["{}".format(i) for i in df.index[np.logical_not(final_mask)]])
        w = f"\nDropping empty lines ({lines})\n\n"
        print(w)

    # Drop rows that are all empty
    df = df.loc[df.index[final_mask],:]

    # -------------------------------------------------------------------------
    # Process required columns. Enforce them as non-null and castable to string

    for c in _private.required_columns:

        try:
            df[c]
        except KeyError:
            err = f"\n\nA topiary dataframe must have a {c} column.\n\n"
            raise ValueError(err)

        # Do a pass trying to infer the datatype of c. (Important in case we
        # dropped empty rows)
        df.loc[:,c] = df.loc[:,c].infer_objects()
        casted = []
        for v in df.loc[:,c]:
            if pd.isna(v) or pd.isnull(v) or v is None or not v or v == "":
                err = f"\nMissing value in required column '{c}'.\n\n"
                raise ValueError(err)

            try:
                casted.append(str(v))
            except (TypeError,ValueError):
                err = f"\nColumn '{c}' must have only strings.\n\n"
                raise ValueError(err)

        df.loc[:,c] = casted

    # -------------------------------------------------------------------------
    # Process sequence column

    # stripping leading/trailing whitespace and removing any in the middle. This
    # is safe because operation above casted to string
    for idx in df.index:
        sequence = str(df.loc[idx,"sequence"]).strip()
        df.loc[idx,"sequence"] = "".join(sequence.split())

    # -------------------------------------------------------------------------
    # Process keep column

    # Try to extract keep column
    try:
        keep = df.loc[:,"keep"]
    # No column: create one and set all to True
    except KeyError:
        keep = None
        df["keep"] = np.ones(len(df),dtype=bool)

    if keep is not None:

        # Do a pass trying to infer the datatype of keep. (This is useful if we
        # dropped empty rows that made the original pandas read this column in
        # as a mix of bool and object).
        df.loc[:,"keep"] = df.keep.infer_objects()

        # If it's not a boolean column, try to turn into one
        if not np.dtype(keep.dtypes) is np.dtype(bool):

            # Base message. If everything works great, let user know what
            # happened as warning. If things go awry, use as start of error
            # message
            w = "\n\n"
            w += "The 'keep' column must be boolean (True/False). pandas did\n"
            w += "not recognize the column as boolean, so we're parsing it\n"
            w += "manually by looking for 0/1, yes/no, true/false, etc.\n\n"

            new_keep = []
            look_for_true = re.compile("[1yt]",re.IGNORECASE)
            look_for_false = re.compile("[0nf]",re.IGNORECASE)
            for k in df.keep:
                if type(k) is bool:
                    is_true = True and k
                    is_false = not is_true
                    looks_like_a = "bool"
                elif type(k) is str:
                    is_true = look_for_true.search(k) is not None
                    is_false = look_for_false.search(k) is not None
                    looks_like_a = "string"
                elif type(k) is int:
                    is_true = (k != 0)
                    is_false = (k == 0)
                    looks_like_a = "int"
                elif type(k) is float:
                    is_true = np.logical_not(np.isclose(k,0))
                    is_false = np.isclose(k,0)
                    looks_like_a = "float"
                else:
                    w += f"Could not figure out how to parse '{k}'\n\n"
                    raise ValueError(w)

                if (is_true and is_false) or (not is_true and not is_false):
                    w += f"Trying to parse '{k}' as a {looks_like_a}, but\n"
                    w += f"could not figure out whether true or false.\n\n"
                    raise ValueError(w)
                else:
                    new_keep.append(is_true)

            # Record newly boolean-ized values
            df.loc[:,"keep"] = np.array(new_keep,dtype=bool)

            # Let user know we manually parsed the keep column...
            print(w)

    # -------------------------------------------------------------------------
    # Process uid column

    warn_uid = ""

    # Make sure the dataframe has a uid column
    try:
        uid = df.loc[:,"uid"]
    except KeyError:
        uid =  None
        warn_uid += "  + No 'uid' column found. Creating column.\n"
        df["uid"] = _private.generate_uid(len(df))

    if uid is not None:

        # Look for a-z and A-Z (only allowed characters in uid)
        p = re.compile("[^a-z]",re.IGNORECASE)

        final_uid = []

        # Go through uid...
        for u in uid:

            # Force to be a spring
            if type(u) is not str:
                u = str(u)

            # Disallowed character(s) in uid or uid not right length
            if p.search(u) or len(u) != 10:

                new_uid = _private.generate_uid()
                warn_uid += f"  + uid '{u}' is invalid. Replacing with '{new_uid}'\n"
                final_uid.append(new_uid)

            else:
                final_uid.append(u)

        df.loc[:,"uid"] = final_uid

    # Make sure uid are unique
    uid, counts = np.unique(df.loc[:,"uid"],return_counts=True)
    for u in uid[counts != 1]:
        warn_uid = f"  + uid '{u}' is not unique. This will be replaced with a\n"
        warn_uid += "    new unique uid for each sequence.\n"

        mask = df.loc[:,"uid"] == u
        df.loc[mask,"uid"] = _private.generate_uid(sum(mask))

    if warn_uid != "":
        w = "\n\n"
        w += "'uid' column was invalid. topiary has fixed the problems noted below.\n"
        w += "If you have already generated phylogenetic trees using a previous\n"
        w += "version of this dataframe, **those trees are no longer compatible\n"
        w += "with this dataframe**. To ensure compatibility, fix the problems\n"
        w += "noted below and then re-read the dataframe into topiary. If you\n"
        w += "have not already generated trees (or plan to generate new ones from\n"
        w += "scratch) you may safely disregard this warning.\n"
        w += "\n"
        w += warn_uid
        print(w)

    # -------------------------------------------------------------------------
    # Check format of ott column if present

    try:
        ott = df.loc[:,"ott"]
    except KeyError:
        ott = None

    if ott is not None:

        for index in df.index:

            # Get ott id
            o = df.loc[index,"ott"]

            # Missing value is okay --> make sure it's a pandas NULL datatype
            if pd.isna(o) or pd.isnull(o) or o is None or not o or o in ["","none","None"]:
                df.loc[index,"ott"] = pd.NA
                continue

            # If there is an ott that is not a null, make sure it is sane and
            # reasable.
            failed = False
            if type(o) is str and o[:3] == "ott":
                try:
                    int(o[3:])
                except ValueError:
                    failed = True
            else:
                failed = True

            if failed:
                err = "\n\nott column must have format 'ottINTEGER', where \n"
                err += "INTEGER is a the integer OTT accession number.\n"
                raise ValueError(err)

    # -------------------------------------------------------------------------
    # Check alignment column
    try:
        alignment = df.loc[:,"alignment"]
    except KeyError:
        alignment = None

    # If there is an alignment column...
    if alignment is not None:

        # Grab non-null alignment lines
        align_mask = np.logical_not(pd.isnull(df.alignment))
        align_mask = np.logical_and(align_mask,df.keep)

        # Make sure alignment column has at least one kept row
        if len(align_mask) > 0 and np.sum(align_mask) > 0:

            # Create matrix holding alignment, rows are sequences, columns are columns
            align_matrix = []
            for row in df.alignment[align_mask]:
                align_matrix.append(list(row))

            # Make sure all alignment rows are the same length
            unique_row_length = set([len(row) for row in align_matrix])
            if len(unique_row_length) != 1:
                err = "\nAll sequences in the 'alignment' column must have the\n"
                err += "same length\n\n"
                raise ValueError(err)

            # Convert to a matrix
            align_matrix = np.array(align_matrix)

            # Create mask for good columns -- columns with more than just "-"
            good_column_mask = []
            for i in range(align_matrix.shape[1]):
                u = np.unique(align_matrix[:,i])
                if len(u) == 1 and u[0] == "-":
                    good_column_mask.append(False)
                else:
                    good_column_mask.append(True)

            good_column_mask = np.array(good_column_mask,dtype=bool)

            # Whack out columns that are only "-"
            align_matrix = align_matrix[:,good_column_mask]

            # Convert alignment matrix back to an array of strings
            new_align = []
            for i in range(align_matrix.shape[0]):
                new_align.append("".join(align_matrix[i,:]))

            # Store in dataframe
            df.loc[align_mask,"alignment"] = new_align

    # -------------------------------------------------------------------------
    # Make sure columns have order nickname, keep, species, name, sequence
    columns = list(df.columns)
    if "nickname" in columns:
        output_order = ["nickname"]
    else:
        output_order = []
    output_order.extend(["keep","species","name","sequence"])

    for c in columns:
        if c not in output_order:
            output_order.append(c)

    return df.loc[:,output_order]



"""
Function to check/process paralog_patterns arguments in topiary functions.
"""

import re, copy

def check_paralog_patterns(paralog_patterns,ignorecase=True,re_flags=None):
    """
    Process a `parlog_patterns` argument and do error checking. Compiles
    regular expressions and returns in a standard format. 

    Parameters
    ----------
    paralog_patterns : dict
        dictionary to check and compile
    ignorecase : bool, default=True
        when compiling regex, whether or not to ignore case
    re_flags : list, optional
        regular expression flags to pass to compile. None or list of
        of flags. Note, "ignorecase" takes precedence over re_flags.

    Returns
    ------
    patterns : list
        list of patterns with format [(compiled_pattern,paralog),...]
    """

    # If nothing is passed in, return a list without throwing an error
    if paralog_patterns is None:
        return []

    # Make generic, informative, error when dealing with paralog_patterns
    generic_pp_error = ["paralog_patterns must be a dictionary keying paralog",
                        "names to patterns that match that paralog. For example,",
                        "the protein LY96 is annotated as LY96, MD2, MD-2,",
                        "ESOP1, Lymphocyte antigen 96, or myeloid differentiation 2.",
                        "A paralog_pattern dictionary for this protein would be",
                        "\nparalog_pattern = {'LY96:['MD2','MD-2','ESOP1',",
                        "                          'lymphocyte antigen 96',",
                        "                          'myeloid differentiation 2']}\n",
                        "The patterns can either be strings (i.e. 'MD-2') or",
                        "re.Pattern instances (i.e. re.compile('MD().*?)2')).",
                        "Note for advanced users, any strings passed in are",
                        "escaped. To use regex, pre-compile your regular",
                        "expressions."]
    generic_pp_error = "\n".join(generic_pp_error)

    # Deal with re_flags argument
    if re_flags is None:
        re_flags = []

    # Parse ignorecase argument
    if ignorecase:
        re_flags.append(re.IGNORECASE)

    # Assemble flags
    re_kwargs = {}
    if len(re_flags) > 0:
        re_kwargs["flags"] = re_flags[0]
        if len(re_flags) > 1:
            for f in re_flags[1:]:
                re_kwargs["flags"] = re_kwargs["flags"] | f

    # Check paralog_patterns data type
    if type(paralog_patterns) is not dict:
        err = "\nparalog_patterns is not a dictionary\n\n{generic_pp_error}\n\n"
        raise ValueError(err)

    # Work on a copy of the paralog_patterns dictionary
    pp = copy.deepcopy(paralog_patterns)

    # Go through all paralog_patterns keys
    patterns = []
    for k in paralog_patterns:

        # Make sure the key is a string
        if type(k) is not str:
            err = f"\nparalog key '{k}' not recognized.\n\n{generic_pp_error}\n\n"
            raise ValueError(err)

        # If value is a naked regular expression pattern or string, wrap it in a
        # list so the code can iterate over it.
        if type(pp[k]) in [re.Pattern,str]:
            pp[k] = [pp[k]]

        # Make sure value is an iterable
        if hasattr(pp[k],"__iter__"):

            # Make sure there is at least one pattern
            if len(pp[k]) == 0:
                err = f"\nparalog key '{k}' has no patterns.\n\n{generic_pp_error}\n\n"
                raise ValueError(err)

            # Compile patterns to search for in the source column
            to_compile = []
            for a in pp[k]:

                if type(a) is str:
                    to_compile.append(re.escape(a))
                elif type(a) is re.Pattern:
                    # NOTE: if user passes multiple compiled patterns with
                    # different flags, this will only use the flags for the
                    # last pattern on *all* regex passed in.
                    to_compile.append(a.pattern)
                    re_kwargs["flags"] = a.flags
                else:
                    err = f"\npattern '{a}' not recognized.\n\n{generic_pp_error}\n\n"
                    raise ValueError(err)

            patterns.append((re.compile("|".join(to_compile),**re_kwargs),k))

        # value is not iterable -- bad news
        else:
            err = f"\nvalue for paralog_pattern '{k}' should be list-like.\n\n"
            err += "{generic_pp_error}\n\n"
            raise ValueError(err)

    return patterns



"""
Construct a topiary dataframe from a seed dataframe.
"""

import topiary
from topiary import check

import numpy as np
import pandas as pd

def df_from_seed(seed_df,
                 ncbi_blast_db="nr",
                 local_blast_db=None,
                 hitlist_size=5000,
                 e_value_cutoff=0.001,
                 gapcosts=(11,1),
                 num_threads=1,
                 **kwargs):
    """
    Construct a topiary dataframe from a seed dataframe, blasting to fill in the
    sequences.

    Parameters
    ----------
    seed_df : pandas.DataFrame or str
        seed dataframe containing seed sequences to launch the analysis. df can
        be a pandas dataframe or a string pointing to a spreadsheet file.
    ncbi_blast_db : str or None, default="nr"
        NCBI blast database to use. If None, use a local database. Incompatible
        with local_blast_db.
    local_blast_db : str or None, default=None
        Local blast database to use. If None, use an NCBI database. Incompatible
        with ncbi_blast_db.
    hitlist_size : int, default=5000
        download only the top hitlist_size hits
    e_value_cutoff : float, default=0.001
        only take hits with e_value better than e_value_cutoff
    gapcost : tuple, default=(11,1)
        BLAST gapcosts (length 2 tuple of ints)
    num_threads : int, default=1
        number of threads to use for BLAST. If -1, use all available threads.
        Note: multithreading rarely speeds up NCBI blast queries, but can
        dramatically speed up local blast searches.
    **kwargs : dict, optional
        extra keyword arguments are passed directly to biopython
        NcbiblastXXXCommandline (for local blast) or qblast (for remote
        blast). These take precedence over anything specified above
        (hitlist_size, for example).

    Returns
    -------
    topiary_dataframe : pandas.DataFrame
        topiary dataframe with sequences found from seed sequence.
    phylo_context : str
        string descriptor of the phylogenetic context on opentreeoflife.org
        (i.e. "Animals" or "All life")
    key_species : list
        list if key species to keep during the analysis
    paralog_patterns : list
        list of compiled regular expressions (extracted from aliases) to use to
        try to match paralogs.

    Notes
    -----
    Every sequence in the original seed dataframe will have `always_keep` set
    to True, so they will not be deleted by subsequent quality control steps.
    """

    out = topiary.io.load_seed_dataframe(seed_df)
    seed_df = out[0]
    phylo_context = out[1]
    key_species = out[2]
    paralog_patterns = out[3]

    # Validate dataframe
    seed_df = check.check_topiary_dataframe(seed_df)

    # Validate blast database arguments
    if ncbi_blast_db is None and local_blast_db is None:
        err = "\nPlease specificy either ncbi_blast_db OR local_blast_db\n\n"
        raise ValueError(err)

    if ncbi_blast_db is not None and local_blast_db is not None:
        err = "\nPlease specificy either ncbi_blast_db OR\n"
        err += "local_blast_db, but not both.\n\n"
        raise ValueError(err)

    # database, hitlist_size, e_value_cutoff, gapcosts, num_threads, and
    # kwargs will all be validated by the blast call itself.

    # ncbi blast
    if ncbi_blast_db:

        # Get the taxid to limit the search to if phylo_context is given
        if phylo_context is not None:
            taxid = topiary.opentree.phylo_to_taxid[phylo_context]
        else:
            taxid = None

        blast_df = topiary.ncbi.ncbi_blast(seed_df.sequence,
                                           db="nr",
                                           taxid=taxid,
                                           hitlist_size=hitlist_size,
                                           e_value_cutoff=e_value_cutoff,
                                           gapcosts=gapcosts,
                                           num_threads=num_threads,
                                           **kwargs)

    # local blast
    if local_blast_db:

        blast_df = topiary.ncbi.local_blast(seed_df.sequence,
                                            db=local_blast_db,
                                            hitlist_size=hitlist_size,
                                            e_value_cutoff=e_value_cutoff,
                                            gapcosts=gapcosts,
                                            num_threads=num_threads,
                                            **kwargs)

    # blast_df is a list of dataframes, one for each hit in seed_df.sequence

    # Go through each blast dataframe
    for i, this_df in enumerate(blast_df):

        # Assign the blast query to a useful name (i.e. LY96|Homo sapiens)
        this_df.loc[:,"query"] = f"{seed_df.name.iloc[i]}|{seed_df.species.iloc[i]}"

        # Parse the blast output from each line to extract the features useful
        # for downstream analyses -- structure, partial, etc.
        # out_dict will be a dictionary keyed to the new columns we want
        # (structure, etc.) with lists of values as long as the dataframe.
        out_dict = None
        for idx in this_df.index:
            parsed = topiary.external.ncbi.parse_ncbi_line(this_df.loc[idx,"title"])
            if out_dict is None:
                out_dict = {}
                for k in parsed:
                    out_dict[k] = [parsed[k]]
            else:
                for k in parsed:
                    out_dict[k].append(parsed[k])

        # Load newly extracted column into the dataframe
        for k in out_dict:
            this_df[k] = out_dict[k]


    df = topiary.ncbi.merge_blast_df(blast_df)

    # Append uid (to prevent user-visible topiary warnings) and then
    # convert to a topiary dataframe
    df["uid"] = topiary._private.generate_uid(len(df.sequence))

    df = topiary.check.check_topiary_dataframe(df)

    df = pd.concat((seed_df,df),ignore_index=True)

    # Set new hits to always_keep
    df.loc[pd.isna(df["always_keep"]),"always_keep"] = False

    # Get ott id for all sequences, setting False for those that can't be
    # found/resolved
    df = topiary.get_ott_id(df,phylo_context=phylo_context,verbose=False)

    # Create nicknames for sequences in dataframe
    df = topiary.create_nicknames(df,paralog_patterns)

    return df, phylo_context, key_species, paralog_patterns



"""
Functions for lowering redundnacy based on clusters of similar sequence
identity.
"""

from Bio import pairwise2

import numpy as np

def _compare_seqs(A_seq,B_seq,A_qual,B_qual,cutoff,discard_key=False):
    """
    Compare sequence A and B based on alignment. If the sequences are
    similar within cutoff, compare A_stats and B_stats and take the sequence
    with the lower score. Scores have left-right priority.  Will select
    sequence with the first element with a lower score. If sequences are
    similar within cutoff and have equal scores, choose A.

    Parameters
    ----------
        A_seq: sequence A
        B_seq: sequence B
        A_qual: quality scores for A
        B_qual: quality scores for B
        cutoff: cutoff for sequence comparison (~seq identity. between 0 and 1)
        discard_key: whether or not to discard key species, regardless of their
                     qualities.

    Return
    ------
        bool, bool

        True, True: keep both
        True, False: keep A
        False, True: keep B
    """

    # Get a normalized score: matches/len(shortest)
    score = pairwise2.align.localxx(A_seq,B_seq,score_only=True)
    norm = score/np.min((len(A_seq),len(B_seq)))

    # If sequence similarity is less than the cutoff, keep both
    if norm <= cutoff:
        return True, True

    # If sequence similarity is greater than the cutoff, select one.
    else:

        # If both always keep, return that we keep both
        if A_qual[0] == 0 and B_qual[0] == 0:
            return True, True

        # If we are not discarding key sequences and both sequences are
        # from key species, automatically keep both.
        if not discard_key:
            if A_qual[1] == 0 and B_qual[1] == 0:
                return True, True

        # Compare two vectors. Identify first element that differs.

        # Return
        #   True, False if A is smaller
        #   False, True if B is smaller
        #   True, False if equal

        comp = np.zeros(A_qual.shape[0],dtype=np.int8)
        comp[B_qual > A_qual] = 1
        comp[B_qual < A_qual] = -1
        diffs = np.nonzero(comp)[0]

        # No difference, keep A arbitrarily
        if diffs.shape[0] == 0:
            return True, False

        # B > A at first difference, keep A
        elif comp[diffs[0]] > 0:
            return True, False

        # B < A at first difference, keep B
        else:
            return False, True

def _check_block_redundancy(args):
    """
    Check for redundancy within a block of the sequence by sequence matrix.
    Updates keep_array in a thread-safe manner. Should be called by
    _reduce_redundancy_thread_manager.


    Parameters
    ----------
        args: iterable of arguments. Expanded to:

        i_block: tuple holding indexes over which to iterate in i
        j_block: tuple holding indexes over which to interate in j
        sequence_array: array holding sequences to compare.
        quality_array: array holding vectors of quality scores, one vector for
                       each sequence
        keep_array: array holding whether or not to keep each sequence. boolean.
                    This array is updated and is the primary output of this
                    function.
        cutoff: float cutoff between 0 and 1 indicating the fractional similarity
                between two sequences above which they are considered redundant.
        discard_key: if discard_key is False, a redundant sequence will be tossed
                     even if it is from a key species
        lock: multiprocessing.Lock instance used to control access to keep_array

    Return
    ------
        None. Updates keep_array in place.
    """

    i_block = args[0]
    j_block = args[1]
    sequence_array = args[2]
    quality_array = args[3]
    keep_array = args[4]
    cutoff = args[5]
    discard_key = args[6]
    lock = args[7]

    # Lock keep_array and make a local copy
    lock.acquire()
    try:
        local_keep = keep_array.copy()
    finally:
        lock.release()

    # Loop over block in i
    for i in range(i_block[0],i_block[1]):

        # Skip if we already know we're not keeping i
        if not local_keep[i]:
            continue

        # Loop over block in j
        for j in range(j_block[0],j_block[1]):

            # Skip if we're below or at diagonal
            if j <= i:
                continue

            # Skip if we already know we're not keeping j
            if not local_keep[j]:
                continue

            # Make comparison
            i_keep, j_keep = _compare_seqs(sequence_array[i],
                                           sequence_array[j],
                                           quality_array[i],
                                           quality_array[j],
                                           cutoff,
                                           discard_key=discard_key)

            # Record results
            local_keep[i] = i_keep
            local_keep[j] = j_keep

            # Not keeping i, we don't need to keep going
            if not i_keep:
                break

    # Lock keep_array and update with what we learned in local_keep. This is
    # an "and" operation. Only keep if keep_array AND local_keep are still True.
    lock.acquire()
    try:
        keep_array[:] = np.logical_and(keep_array,local_keep)
    finally:
        lock.release()



"""
A python framework for doing ancestral sequence reconstruction using
pandas dataframes and ete3 trees as the primary data structures.
"""
__author__ = "Michael J. Harms"

# Submodules
from . import util
from . import draw
from . import pipeline
from . import _private
#from . import check
from . import pipeline

from .external import generax
from .external import ncbi
from .external import muscle
from .external import opentree
from .external import raxml

# Core functions for pipeline
from .quality import remove_redundancy, clean_alignment, taxonomic_sample
from .pipeline import seed_to_alignment
from .util import create_nicknames
from .external import get_ott_id, get_species_tree
from .external import recip_blast
from .external import run_muscle
from .external import find_best_model, generate_ml_tree, generate_ancestors
from .external import reconcile

# Input/output functions
from .io import df_from_blast_xml, df_from_seed
from .io import write_dataframe, read_dataframe
from .io import read_fasta_into, write_fasta, write_phy

# Topiary version
from .__version__ import __version__

def _check_for_notebook():
    """
    Check whether the code is being executed in a notebook or standard
    standard python interpreter.

    Return
    ------
        string for jupyter or IPython, None for something not recognized.
    """

    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return "jupyter"   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return "IPython"   # Terminal running IPython
        else:
            return None        # Not sure what interpreter

    # Probably standard Python interpreter
    except NameError:
        return None

_in_notebook = _check_for_notebook()



"""
Get OTT ids given a topiary dataframe.
"""

import topiary
from topiary import check
from topiary.external.opentree.util import species_to_ott

from opentree import OT, taxonomy_helpers
import dendropy as dp
import ete3

import pandas as pd
import numpy as np

import re, copy

def get_ott_id(df,
               phylo_context="All life",
               verbose=True):
    """
    Return a copy of df with an ott column holding open tree of life
    names for each species.

    Parameters
    ----------
    df : pandas.DataFrame
        dataframe that has an ott column with Open Tree of Life taxon ids
    phylo_context: str, default="All life"
        used to limit scope for looking up species. See Notes for details.
    verbose : bool, default=True
        whether or not to print out unresolvable taxa, etc.

    Returns
    -------
    topiary_df : pandas.DataFrame
        Copy of df with added ott and orig_species column.  ott column
        holds ott index for the species. orig_species holds what used to be
        in the species column. The species column is replaced by the uniuqe
        species name used by Open Tree of Life.

    Notes
    -----
    `phylo_context` should be a string recognized by opentreeoflife.org. To get
    the current strings recognized by the database, use the following code:

    .. code-block:: python

        from opentree import OT
        print(OT.tnrs_contexts().response_dict)

    As of 2022-06-21, the following are recognized. You can use
    either the keys or values in this dictionary as a `phylo_context`.

    .. code-block:: python

        {'ANIMALS':
            ['Animals', 'Birds', 'Tetrapods', 'Mammals', 'Amphibians',
             'Vertebrates', 'Arthropods', 'Molluscs', 'Nematodes',
             'Platyhelminthes', 'Annelids', 'Cnidarians', 'Arachnids','Insects'],
         'FUNGI': ['Fungi', 'Basidiomycetes', 'Ascomycetes'],
         'LIFE': ['All life'],
         'MICROBES': ['Bacteria', 'SAR group', 'Archaea', 'Excavata', 'Amoebozoa',
                      'Centrohelida', 'Haptophyta', 'Apusozoa', 'Diatoms',
                      'Ciliates', 'Forams'],
         'PLANTS': ['Land plants', 'Hornworts', 'Mosses', 'Liverworts',
                    'Vascular plants', 'Club mosses', 'Ferns', 'Seed plants',
                    'Flowering plants', 'Monocots', 'Eudicots', 'Rosids',
                    'Asterids', 'Asterales', 'Asteraceae', 'Aster',
                    'Symphyotrichum', 'Campanulaceae', 'Lobelia']}

    """

    # Make sure this is a topiary dataframe
    df = check.check_topiary_dataframe(df)

    # Make sure verbose can be properly handled
    verbose = check.check_bool(verbose,"verbose")

    # Make copy of df and copy current species to original species
    local_df = df.copy()
    local_df["orig_species"] = local_df.loc[:,"species"]

    # Get unique list of species, stripping any leading/trailing spaces.
    species_list = list(local_df.species.drop_duplicates())

    # Actually get species ott from opentree database
    results, not_resolved = species_to_ott(species_list=species_list,
                                           phylo_context=phylo_context)

    # Create new, empty column for "ott" in the local df
    local_df["ott"] = pd.array(["" for _ in range(len(local_df))])

    # Go through the local_df and populate species, ott, and keep
    unrecognized_name = []
    unresolved_taxa = []
    final_ott = []
    for i in range(len(local_df)):

        # Get row name
        row_name = local_df.iloc[i].name

        # Get species and what keep status was before this move
        s = local_df.loc[row_name,"orig_species"]
        keep = local_df.loc[row_name,"keep"]

        # Default values
        ott_id = None
        species = s

        # Try to grab parsed results. If fails, no hit. Set ott_id to None
        # and keep species as is
        try:
            ott_id = results[s][0]
            species = results[s][1]
        except KeyError:
            pass

        # If ott is none, set to keep = False, and record we did not find the
        # species
        if ott_id is None:
            keep = False
            unrecognized_name.append(s)

        # If ott_id was something not resovled in the syntehci tree, set
        # keep = False and record we could not resolve
        if ott_id in not_resolved:
            keep = False
            unresolved_taxa.append(s)

        # Update the local_df keep, species, and ott
        local_df.loc[row_name,"keep"] = keep
        local_df.loc[row_name,"species"] = species

        if ott_id is None:
            local_df.loc[row_name,"ott"] = pd.NA
        else:
            local_df.loc[row_name,"ott"] = f"ott{ott_id}"

    # Print warning data for user -- species we could not find OTT for
    unrecognized_name = set(unrecognized_name)
    if len(unrecognized_name) != 0:

        w = "\n"
        w += "Could not find OTT for following species:\n"
        for u in unrecognized_name:
            w += f"    {u}\n"

        w += "\nSetting `keep = False` for all of these species\n"
        w += re.sub("        ","",
        """
        topiary looks up unique identifiers for every species (OTT ids) using the
        opentreeoflife database. This did not work for the species listed above.
        For the moment, topiary has simply set `keep = False` for any sequences
        from these species in the dataframe, meaning they will be excluded from
        the analysis. If you want to keep these sequences, you can look up the
        OTT for the species manually on https://tree.opentreeoflife.org/,
        manually add that ott to the dataframe, and set `keep = True` for that
        row.

        This is often caused when a species has two names (for example,
        Apteryx mantelli mantelli vs. Apteryx australis mantelli). If ncbi uses
        one species name and opentreeoflife uses another, this will lead to this
        error.  This can also occur when the ncbi species name is ambiguous,
        referring to genus/species pair that has more than one subspecies
        annotated in opentreeoflife. A final common problem is when the ncbi
        sequence comes from a hybrid (for example, Bos indicus x Bos taurus).
        This is a unique species, but can't be placed on a bifurcating species
        tree.

        If you are able to find a name for the spieces that successfully resolves
        on the opentreeoflife database, you can update the dataframe. For the
        example of Apteryx mantelli mantellii above, you could fix this error
        by running the following code. (Note we set `keep = True` because the
        failed look up automatically set `keep = False` fo these species.)

        ```
        df.loc[df.species == "Apteryx mantelli mantelli","species"] = "Apteryx australis mantelli"
        df.loc[df.species == "Apteryx australis mantelli","keep"] = True
        df = topiary.get_ott_id(df,phylo_context="Animals")
        ```
        \n""")

        if verbose:
            print(w)

    # Print warning data for user -- species we could not resolve
    unresolved_taxa = set(unresolved_taxa)
    if len(unresolved_taxa) != 0:

        w = "\n"
        w += "Following species have OTT, but cannot be placed on tree:\n"
        for u in unresolved_taxa:
            w += f"    {u}\n"
        w += "\nSetting `keep = False` for all of these species.\n"
        w += re.sub("        ","",
        """
        topiary looks up unique identifiers for every species (OTT ids) using the
        opentreeoflife database. This did not work for the species listed above.
        For the moment, topiary has simply set `keep = False` for any sequences
        from these species in the dataframe, meaning they will be excluded from
        the analysis.

        This particular problem usually occurs for hybrids (i.e. mules). These
        cannot be placed on a branching tree because they are the result of a
        cross between branches. You have two options to fix this problem. 1)
        Remove the sequence from the analysis. This is what will happen if you
        leave `keep = False`. 2) If you need this sequence, look up the OTT for
        one of the hybrid parents (e.g. horse/donkey for mule) on
        https://tree.opentreeoflife.org/ and then use that as the OTT for this
        sequence. This is only recommended if you do not have the parent species
        in the alignment already. After adding the new OTT, make sure to set
        `keep = True` for this sequence.
        """)

        if verbose:
            print(w)

    return local_df



"""
Generic interface for wrapping external programs.
"""

import topiary

import pandas as pd

import subprocess, os, sys, time, random, string, shutil, copy, json
import multiprocessing as mp

def gen_seed():
    """
    Generate a random string of 10 integers and return as a string.

    Return
    ------
        10-digit integer as a string
    """

    return "".join([f"{random.choice(range(10)):d}" for _ in range(10)])

def create_new_dir(dir_name=None,overwrite=False):
    """
    Create a new directory.

    Parameters
    ----------
        dir_name: if specified, name the directory this
        overwrite: overwrite existing directory

    Return
    ------
        name of created directory
    """

    # if dir_name is not specified, build it in a stereotyped fashion
    if dir_name is None:
        rand_name = "".join([random.choice(string.ascii_letters)
                              for _ in range(10)])
        dir_base = "calculation"

        dir_name = f"{dir_base}_{rand_name}"

    # If directory already exists, throw error
    if os.path.exists(dir_name):
        if overwrite:
            shutil.rmtree(dir_name)
        else:
            err = f"{dir_name} already exists.\n"
            raise FileExistsError(err)

    # Make directory
    os.mkdir(dir_name)

    return os.path.abspath(dir_name)

def copy_input_file(input_file,
                    dir_name,
                    file_name=None,
                    put_in_input_dir=True):
    """
    Copy an input file into a directory in a stereotyped way.

    If put_in_input_dir is specified, copy input_file into dir_name/input,
    creating input if necessary.  If put_in_input_dir is not specified,
    copy in the file as dir_name/{input_file}.

    Parameters
    ----------
        input_file: file to copy in
        dir_name: copy into dir_name
        file_name: what to call file in new directory. If None, use same name.
        put_in_input_dir: (bool) copy into directory input or not.

    Return
    ------
        name of copied file
    """

    if not os.path.isfile(input_file):
        err = f"input_file {input_file} does not exist\n"
        raise FileNotFoundError(err)

    if not os.path.isdir(dir_name):
        err = f"dir_name {dir_name} is not a directory\n"
        raise FileNotFoundError(err)

    if file_name is None:
        file_name = os.path.split(input_file)[-1]
    file_alone = os.path.split(file_name)[-1]

    # If we are putting this into an input subdirectory...
    if put_in_input_dir:
        input_dir = os.path.join(dir_name,"input")
        if not os.path.exists(input_dir):
            os.mkdir(input_dir)
        file_alone = os.path.join("input",file_alone)

    # Copy in file
    out_file = os.path.join(dir_name,file_alone)
    shutil.copy(input_file,out_file)

    return os.path.abspath(out_file)

def read_previous_run_dir(previous_dir):
    """
    Load the df, tree, and run_parameters from a previous run directory.

    Parameters
    ----------
        previous_dir: directory of previous run (top level; meaning, ml-tree
                      not ml-tree/output)

    Return
    ------
        dictionary with df, tree_file, and parameters
    """

    previous = {}

    # Make sure previous_dir is a string.
    if type(previous_dir) is not str:
        err = f"\nprevious_dir '{previous_dir}' not recognized. Should be a string.\n"
        raise ValueError(err)

    # Look for output directory
    out_dir = os.path.join(previous_dir,"output")
    if not os.path.isdir(out_dir):
        err = f"\nCould not read previous directory '{previous_dir}'. This\n"
        err += "should be a previous topiary run directory that contains an\n"
        err += "'output' directory.\n"
        raise FileNotFoundError(err)

    # Grab the run parameters
    try:
        f = open(os.path.abspath(os.path.join(out_dir,"run_parameters.json")),'r')
        run_parameters = json.load(f)
        f.close()
    except FileNotFoundError as e:
        err = f"\nCould not read previous directory '{previous_dir}'. This\n"
        err += "directory should contain a file output/run_parameters.json.\n"
        err += "This file was not found.\n\n"
        raise FileNotFoundError(err) from e
    except json.JSONDecodeError as e:
        err = f"\nCould not read previous directory '{previous_dir}'. This\n"
        err += "directory should contain a file output/run_parameters.json.\n"
        err += "This file exists but could not be read!\n\n"
        raise ValueError(err) from e

    # copy in previous run parameters
    for k in run_parameters:
        previous[k] = run_parameters[k]

    # Try to grab dataframe
    df_file = os.path.abspath(os.path.join(out_dir,"dataframe.csv"))
    if os.path.exists(df_file):
        previous["df"] = topiary.read_dataframe(df_file)
    else:
        err = f"\nCould not read previous directory '{previous_dir}'. Does not\n"
        err += "contain the file output/dataframe.csv"
        raise FileNotFoundError(err)

    # Try to grab the tree file
    tree_file = os.path.abspath(os.path.join(out_dir,"tree.newick"))
    if os.path.exists(tree_file):
        previous["tree_file"] = tree_file

    return previous


def prep_calc(previous_dir=None,
              df=None,
              model=None,
              tree_file=None,
              other_files=[],
              output=None,
              overwrite=False,
              output_base="calculation"):
    """
    Prepare a calculation, organizing input files and creating an output
    directory.

    Parameters
    ----------
        previous_dir: directory containing previous calculation. prep_calc will
                      grab the the csv, model, and tree from the previous run.
        These arguments will override anything pulled out from the previous_dir.
        df: topiary data frame or .csv file written out from a topiary data frame
        model: phylogenetic model as string (e.g. JTT, LG+G8)
        tree_file: newick tree file
        other_files: other files besides the df and tree that are needed
        output: output directory. If not specified, create an output directory with
                form "{output_base}_randomletters"
        overwrite: whether or not to overwrite existing (default is False)
        output_base: base to assign the directory if no output is specified.

    Return
    ------
        dictionary containing information about the prepared calculation
        (things like filenames, etc.)
    """

    # -------------------------------------------------------------------------
    # Load in information from the previous calculation

    previous = {}
    if previous_dir is not None:
        previous = read_previous_run_dir(previous_dir)

    # -------------------------------------------------------------------------
    # Create output directory

    output = create_new_dir(dir_name=output,overwrite=overwrite)

    # -------------------------------------------------------------------------
    # df (dataframe)

    # If passed df is None, try to grab previous.
    if df is None:
        try:
            df = previous["df"]
        except KeyError:
            pass

    # Read dataframe
    csv_file = None
    if df is not None:

        # Read dataframe
        if type(df) is pd.DataFrame:
            df = topiary.check.check_topiary_dataframe(df)

        # Read csv
        elif type(df) is str:
            csv_file = copy.deepcopy(df)
            df = topiary.read_dataframe(csv_file)
        else:
            err = "df must be a pandas data frame or csv file that can be read as\n"
            err += "a pandas data frame\n"
            raise ValueError(err)

    if df is None:
        err = "A dataframe must be specified, either explicitly via the df \n"
        err += "argument or implicitly by having dataframe.csv already present\n"
        err += "in prev_dir.\n"
        raise ValueError(err)

    # -------------------------------------------------------------------------
    # Model

    # If passed model is None, try to grab previous
    if model is None:
        try:
            model = previous["model"]
        except KeyError:
            pass

    if model is not None:
        if type(model) is not str:
            err = f"\nmodel '{model}' not recognized. Should be a string.\n"
            raise ValueError(err)

    # -------------------------------------------------------------------------
    # Tree file

    # If passed tree_file is None, try to grab previous
    if tree_file is None:
        try:
            tree_file = previous["tree_file"]
        except KeyError:
            pass

    # Copy in tree file
    if tree_file is not None:
        tree_file = copy_input_file(tree_file,output,put_in_input_dir=True)

    # -------------------------------------------------------------------------
    # Copy remaining files into output/input directory

    final_files = []
    for f in other_files:
        if f is None:
            final_files.append(None)
        else:
            final_files.append(copy_input_file(f,
                                               output,
                                               put_in_input_dir=True))

    # Move into the output directory
    starting_dir = os.getcwd()
    os.chdir(output)

    # Make sure we actually generated the input directory at some point in the
    # steps above
    if not os.path.exists("input"):
        os.mkdir("input")

    # Write the alignment to the input directory
    alignment_file = os.path.abspath(os.path.join("input","alignment.phy"))
    topiary.write_phy(df,out_file=alignment_file,seq_column="alignment",
                      write_only_keepers=True,clean_sequence=True)

    df_file = os.path.join("input","dataframe.csv")
    topiary.write_dataframe(df,out_file=df_file)

    # Populate output with data pulled from previous run
    out = {}
    for k in previous:
        out[k] = previous[k]

    # Override with results of function parsing above.
    out["df"] = df
    out["csv_file"] = csv_file
    out["tree_file"] = tree_file
    out["model"] = model
    out["alignment_file"] = alignment_file
    out["previous_dir"] = previous_dir
    out["starting_dir"] = starting_dir
    out["output"] = output
    out["other_files"] = final_files

    return out

def _subproc_wrapper(cmd,stdout,queue):
    """
    Wrap the subprocess.run call to allow multithreading.

    Parameters
    ----------
        args: args to pass to subprocess.run
        kwargs: kwargs to pass to subprocess.run
        queue: multiprocessing queue to catch return value

    Return
    ------
        None
    """

    ret = subprocess.run(cmd,stdout=stdout)
    queue.put(ret)

def _follow_log_generator(f,queue):
    """
    Generator function that follows some file object (f) until a
    multiprocessing Queue (queue) is not empty. This is useful for following a
    log file being spit out by an external program whose return will be put
    into the queue.

    Parameters
    ----------
        f: open file object
        queue: multiprocessing.Queue object

    Return
    ------
        None
    """

    # start infinite loop
    found_line = False
    while queue.empty():

        # read last line of file
        line = f.readline()

        # sleep if file hasn't been updated
        if not line:
            time.sleep(0.1)
            continue

        found_line = True

        yield line

    # If we found a line at one point and the job has finished, keep reading
    # lines until we run out -- that is, get trailing data written to log file
    # after job is done
    counter = 0
    if found_line and not queue.empty():
        while True:
            counter += 1
            line = f.readline()
            if line is None or line == "":
                return

            yield line

            if counter > 200:
                break

def launch(cmd,run_directory,log_file=None):
    """
    Launch an external command in a specific directory. Runs command on its own
    thread, allowing python to capture output from a file as standard output.

    Parameters
    ----------
        cmd: subprocess style command list (i.e. ["ls","-al"])
        run_directory: directory in which to run the command. changes into that
                       directory, then returns out when process completes
        log_file: log file where command output will be stored. if specified, the
                  output of the log file is captured and written to standard output
                  (equivalent to `tail -f log_file`).

    Return
    ------
        None. Throws RuntimeError if the command terminates unexpectedly.
    """

    # Go into working directory
    cwd = os.getcwd()
    os.chdir(run_directory)

    # Print command
    full_cmd = " ".join(cmd)
    print(f"Running '{full_cmd}'",flush=True)

    # Launch as a multiprocessing process that will return its output to a
    # multiprocessing queue.
    queue = mp.Queue()
    main_process = mp.Process(target=_subproc_wrapper,
                              args=(cmd,subprocess.PIPE,queue))
    main_process.start()

    # If following a log
    if log_file is not None:

        # If queue is not empty, the job has finished and put its return value
        # into the queue
        while queue.empty():

            # Try to open log every second
            try:
                f = open(log_file,"r")

                # Use follow generator function to catch lines as the come out
                for line in _follow_log_generator(f,queue):
                    print(line,flush=True,end="")
                f.close()

            except FileNotFoundError:
                time.sleep(1)

    # Wait for main process to complete and get return
    ret = queue.get()
    main_process.join()

    # Check for error on return
    if ret.returncode != 0:
        err = f"ERROR: {cmd[0]} returned {ret.returncode}\n\n"
        err += "------------------------------------------------------------\n"
        err += f" {cmd[0]} output \n"
        err += "------------------------------------------------------------\n"
        err += "\n\n"

        err += "".join([line for line in ret.stdout.decode()])

        raise RuntimeError(err)

    # Leave working directory
    os.chdir(cwd)

def write_run_information(outdir,df,calc_type,model,cmd,outgroup=None):
    """
    Write information from the run in a standard way.

    Parameters
    ----------
        outdir: output directory (i.e. ml-run/output)
        df: dataframe to write to dataframe.csv
        calc_type: calculation type (string)
        model: model (string)
        cmd: command invoked to raxml or generax (string)
        outgroup: length 2 list with two sets of outgroups

    Return
    ------
        None
    """

    # Dataframe
    topiary.write_dataframe(df,os.path.join(outdir,"dataframe.csv"))

    # Run parameters
    out_dict = {"calc_type":calc_type,
                "model":model,
                "cmd":cmd,
                "version":topiary.__version__,
                "outgroup":outgroup}

    f = open(os.path.join(outdir,"run_parameters.json"),"w")
    json.dump(out_dict,f)
    f.close()



"""
Merge dataframes from multiple BLAST queries.
"""

import topiary
from topiary import check

import numpy as np
import pandas as pd

def _check_merge(merge_this,merge_into,merge_list):
    """
    Merge the row given by index into the row given by merge_index. If the row
    given by merge_index is itself already merging, merge the index row into
    the existing merge row.

    Parameters
    ----------
        index: index of the row we are going to merge
        merge_index: index of the row we are going to merge INTO
        to_merge: list of sequences to merge. if to_merge[merge_index] is
                  already defined, set merge_index -> to_merge[merge_index]
                  and then do to_merge[index] = merge_index

    Return
    ------
        updated to_merge list
    """

    # If merge_list[merge_into] is not defined, set it to merge_into. If it is
    # defined, set merge_into to merge_list[merge_into]
    if merge_list[merge_into] is None:
        merge_list[merge_into] = merge_into
    else:
        merge_into = merge_list[merge_into]

    # If the sequence we're merging has not been merged, just record.
    if merge_list[merge_this] is None:
        merge_list[merge_this] = merge_into
    else:

        # If the current value in merge_list is *not* merge_into, update all
        # entries that have the current value to be merge_into.
        if merge_list[merge_this] != merge_into:
            curr_value = merge_list[merge_this]
            for i in range(len(merge_list)):
                if merge_list[i] in [curr_value,merge_into]:
                    merge_list[i] = merge_into

    return merge_list

def merge_blast_df(blast_df_list):
    """
    Merge dataframes from multiple BLAST queries. Merge happens based on
    accession. If different queries returned an overlapping sequence region
    from the same accession, merge them together. The merge can include an
    arbitrary number of sequences. This merge will:

    1. Take the longest contiguous hit as the row to keep.
    2. Update the "query" column to include all merged queries. For
       example, if the Human and Troll queries gave overlapping hits from
       subject region 1-100 and 2-101, respectively, the merged query will
       be "Human|1-100;Troll|2-101.
    3. Expand the "subject_start" and "subject_end" numbers to cover the
       maximum extent of all hits. In the example from (1), these would
       become 1 and 101 (human start, troll end).
    4. Set the "e_value" column to be the lowest e_value observed by any
       of the merged hits.

    Parameters
    ----------
    blast_df_list : list
        list of dataframes returned by a blast call with multiple sequence
        queries

    Returns
    -------
    merged : pandas.DataFrame
        single dataframe with duplicate hits between queries merged
    """

    # Make sure blast_df_list is iterable
    try:
        blast_df_list = check.check_iter(blast_df_list,
                                                     "blast_df_list")
    except ValueError:
        err = "\nblast_df_list must be a list of pandas dataframes\n\n"
        raise ValueError(err)

    # Validate input
    required_columns = ["accession","subject_start","subject_end","query","e_value"]
    for d in blast_df_list:

        # Make sure elements are all dataframes
        if type(d) is not type(pd.DataFrame()):
            err = "\nblast_df_list must be a list of pandas dataframes\n\n"
            raise ValueError(err)

        # Make sure the dataframe has the required columns
        for r in required_columns:
            if not r in d.columns:
                err = f"\ndataframe in blast_df_list does not have required column '{r}'.\n"
                err += "Dataframes must have columns:\n"
                for c in required_columns:
                    err += f"    {c}\n"
                err += "\n"
                raise ValueError(err)

    # Don't do anything if we send in a single entry -- just return a
    # copy of the dataframe. (We send out a copy because the merged dataframe
    # will generally not be the same object as the inputs).
    if len(blast_df_list) == 0:
        return blast_df_list[0].copy()

    # concatenate dataframes
    df = pd.concat(blast_df_list,ignore_index=True)

    # Mask will hold whether to keep each row or not
    keep_mask = np.ones(len(df.index),dtype=bool)

    # Go over all accessions
    accessions = np.unique(df.loc[:,"accession"])
    for a in accessions:

        # Grab every row with this accession
        mask = df.accession == a

        # If only one row, move on
        num_entries = np.sum(mask)
        if num_entries == 1:
            continue

        # Get copy of dataframe with only this accession
        this_df = df.loc[mask,:]

        # Compare all entries that have the same accession
        to_merge = [None for _ in range(num_entries)]
        for i in range(num_entries):

            i_start = this_df.subject_start.iloc[i]
            i_end = this_df.subject_end.iloc[i]

            for j in range(i+1,num_entries):

                j_start = this_df.subject_start.iloc[j]
                j_end = this_df.subject_end.iloc[j]

                # If they overlap, merge them
                if not (j_end < i_start or i_end < j_start):
                    to_merge = _check_merge(i,j,to_merge)

        # Get unique overlapping seuqneces
        clusters = list(set(to_merge))
        for cluster in clusters:

            # Don't merge; no overlaps detected
            if cluster is None:
                continue

            # Get all rows to merge, including the one to keep
            merge_idx = np.array([cluster == c for c in to_merge],dtype=bool)
            rows_to_merge = this_df.index[merge_idx]
            keep_mask[rows_to_merge] = False

            # Get the one row to keep
            row_to_keep = this_df.index[cluster]
            keep_mask[row_to_keep] = True

            # Create new start, ends, querie, and e-value for row to keep

            # Get starts, ends, queries, and e-value for the rows to merge
            starts = this_df.loc[rows_to_merge,"subject_start"]
            ends = this_df.loc[rows_to_merge,"subject_end"]
            queries = this_df.loc[rows_to_merge,"query"]
            e_value = this_df.loc[rows_to_merge,"e_value"]

            # Broadest start to end from the merged set
            new_start = np.min(starts)
            new_end = np.max(ends)

            # Lowest e-value from merged set
            new_evalue = np.min(e_value)

            # Create a new query, merging the queries and starts/stops from the
            # merged rows.
            new_query = []
            for i in range(len(queries)):
                new_query.append(f"{queries.iloc[i]}|{starts.iloc[i]}-{ends.iloc[i]}")
            new_query = ";".join(new_query)

            # Update the row to keep with the new query, start, and end
            df.loc[row_to_keep,"query"] = new_query
            df.loc[row_to_keep,"subject_start"] = new_start
            df.loc[row_to_keep,"subject_end"] = new_end
            df.loc[row_to_keep,"e_value"] = new_evalue


    return df.loc[keep_mask,:].reset_index().drop(columns=["index"])



"""
Remove redundancy for datasets in a semi-intelligent way.
"""

import topiary
from topiary import check

from ._block import _check_block_redundancy

import pandas as pd
import numpy as np
from tqdm.auto import tqdm

import os
import multiprocessing as mp

# Columns to check in order
_EXPECTED_COLUMNS = ["structure","low_quality","partial","predicted",
                     "precursor","hypothetical","isoform","diff_from_median"]

class _DummyTqdm():
    """
    Fake tqdm progress bar so we don't have to show a status bar if we don't
    want to. Can be substituted wherever we would use tqdm (i.e.
    tqdm(range(10)) --> DummyTqdm(range(10)).
    """

    def __init__(self,*args,**kwargs):
        pass
    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        pass

    def update(self,value):
        pass

class _FakeLock():
    """
    Fake multiprocessing.Lock instance.
    """
    def __init__(self):
        pass

    def acquire(self):
        pass

    def release(self):
        pass

def _get_quality_scores(row,key_species={}):
    """
    Get stats in order of importance for a specific sequence. (see
    remove_redundancy doc string).

    Parameters
    ----------
        row: row from dataframe built by topiary
        key_species: dictionary of key species to prefer to others. only uses
                     keys for fast look up and ignores values

    Return
    ------
        float array for the sequence
    """

    try:
        if row.always_keep:
            values = [0]
        else:
            values = [1]
    except AttributeError:
        values = [1]

    # See if this is a key species
    try:
        key_species[row.species]
        values.append(0)
    except KeyError:
        values.append(1)

    # Add values for expected columns
    values.extend(list(row[_EXPECTED_COLUMNS]))

    # Flip length column
    values.append(1/len(row.sequence))

    return np.array(values,dtype=float)

def _reduce_redundancy_thread_manager(sequence_array,
                                      quality_array,
                                      keep_array,
                                      cutoff,
                                      discard_key,
                                      num_threads=-1,
                                      progress_bar=True):
    """
    Break sequence_array into a rational number of blocks given the number of
    threads and then run _check_block_redundancy on the sequence array on
    multiple threads. Updates keep_array in place.

    Parameters
    ----------
        sequence_array: array holding sequences to compare.
        quality_array: array holding vectors of quality scores, one vector for
                       each sequence
        keep_array: array holding whether or not to keep each sequence. boolean.
                    This array is updated and is the primary output of this
                    function.
        cutoff: float cutoff between 0 and 1 indicating the fractional similarity
                between two sequences above which they are considered redundant.
        discard_key: if discard_key is False, a redundant sequence will be tossed
                     even if it is from a key species
        num_threads: number of threads to use. if -1 use all available
        progress_bar: whether or not to show a progress bar

    Return
    ------
        num_threads and all_args. This is for testing/debugging purposes.
        Usually ignored. Updates keep_array in place -- this is the main output.
    """


    # Try to figure out how many cores are available
    try:
        num_machine_threads = mp.cpu_count()
    except NotImplementedError:
        num_machine_threads = os.cpu_count()

    # If we can't figure it out, revert to 1
    if num_machine_threads is None:
        print("Could not determine number of cpus. Using single thread.\n")
        num_machine_threads = 1

    # Determine number of threads useful for this problem. It's not worth
    # chopping up a super small set of comparisons
    max_useful_threads = len(sequence_array)//50
    if max_useful_threads < 1:
        max_useful_threads = 1
    if max_useful_threads > num_machine_threads:
        max_useful_threads = num_machine_threads

    # Set number of threads
    if num_threads == -1 or num_threads > max_useful_threads:
        num_threads = max_useful_threads

    # If only using one thread, don't waste overhead of making pool
    if num_threads == 1:
        block = (0,len(sequence_array))
        args = (block,block,
                sequence_array,quality_array,keep_array,
                cutoff,discard_key,_FakeLock())
        _check_block_redundancy(args)
        return 1, (args,)


    num_blocks = (num_threads**2 - num_threads)//2 + num_threads
    block_size = len(sequence_array)//num_threads

    print(f"Calculating redundancy for {num_blocks} blocks of ~{block_size} x {block_size} sequences.",flush=True)

    # Lock allowing threads to safely update keep_array
    manager = mp.Manager()
    lock =  manager.Lock()
    with mp.Pool(num_threads) as pool:

        # Calculate window sizes to cover whole L x L array, where L is number
        # of sequences
        windows = np.zeros(num_threads,dtype=int)
        windows[:] = len(sequence_array)//num_threads

        # This call spreads remainder of L/num_threads evenly across first
        # remainder windows
        windows[:(len(sequence_array) % num_threads)] += 1

        # Blocks will allow us to tile over whole redundancy matrix
        all_args = []
        for i in range(num_threads):
            for j in range(i,num_threads):

                i_block = (np.sum(windows[:i]),np.sum(windows[:i+1]))
                j_block = (np.sum(windows[:j]),np.sum(windows[:j+1]))

                all_args.append((i_block,
                                 j_block,
                                 sequence_array,
                                 quality_array,
                                 keep_array,
                                 cutoff,
                                 discard_key,
                                 lock))

        # Black magic. pool.imap() runs a function on elements in iterable,
        # filling threads as each job finishes. (Calls _check_block_redundancy
        # on every args tuple in all_args). tqdm gives us a status bar.
        # By wrapping pool.imap iterator in tqdm, we get a status bar that
        # updates as each thread finishes.
        if progress_bar:
            list(tqdm(pool.imap(_check_block_redundancy,all_args),total=len(all_args)))
        else:
            pool.imap(_check_block_redundancy,all_args)

    return num_threads, all_args


def remove_redundancy(df,
                      cutoff=0.95,
                      key_species=[],
                      silent=False,
                      only_in_species=False,
                      num_threads=-1):
    """
    Remove redundant sequences according to cutoff and semi-intelligent
    heuristics.

    Favors sequences according to the following criteria, in order:

    1. whether sequence is from a key species
    2. whether it's annotated as structure
    3. how different the sequence length is from the median length
    4. whether it's annotated as low quality
    5. whether it's annotated as partial
    6. whether it's annotated as precursor
    7. whether it's annotated as hypothetical
    8. whether it's annotated as isoform
    9. sequence length (preferring longer)

    Parameters
    ----------
    df : pandas.DataFrame
        topiary data frame with sequences
    cutoff : float, default=0.95
        %identity cutoff for combining removing sequences (between 0 and 1)
    key_species : list, default=[]
        key species to preferentially keep
    silent : bool, default=False
        whether to print output and use status bars
    only_in_species : bool, default=False
        only reduce redundancy within species; do not compare sequences between
        species
    num_threads : int, default=-1
        number of threads to use. If -1, use all available

    Returns
    -------
    topiary_dataframe : pandas.dataframe
        Copy of df in which "keep" is set to False for redundant sequences
    """

    # Process arguments
    df = check.check_topiary_dataframe(df)
    cutoff = check.check_float(cutoff,
                                           "cutoff",
                                           minimum_allowed=0,
                                           maximum_allowed=1)
    key_species = check.check_iter(key_species,
                                               "key_species",
                                               required_value_type=str,
                                               is_not_type=[str,dict])
    silent = check.check_bool(silent,"silent")

    # Encode key species as a dictionary for fast look up
    key_species = dict([(k,None) for k in key_species])

    starting_keep_number = np.sum(df.keep)

    # If not more than one seq, don't do anything
    if len(df) < 2:
        return df

    # Make sure the dataframe has the columns needed for this comparison. If
    # the dataframe does not have the column, simply set to False
    for e in _EXPECTED_COLUMNS:

        try:
            v = df[e]
            v*1.0

        # Column doesn't exist -- record as False
        except KeyError:
            df[e] = False

        # Column exists but can't be interpreted as float. Throw error.
        except TypeError:
            err = "\nThe remove_redundancy function expects a dataframe with \n"
            err += f"column '{e}' that can be interpreted as a number. This\n"
            err += f"column exists but has datatype '{df.loc[:,e].dtype}'.\n"
            err += "To fix, please rename this column and re-run this function.\n\n"
            raise ValueError(err)

    # Figure out how different each sequence is from the median length.  We
    # want to favor sequences that are closer to the median length than
    # otherwise.
    lengths = np.array([len(s) for s in df.loc[:,"sequence"]],dtype=int)
    kept_lengths = lengths[df.keep]
    hist_counts, hist_lengths = np.histogram(kept_lengths,
                                             bins=int(np.round(2*np.sqrt(len(kept_lengths)),0)))
    median_length = hist_lengths[np.argmax(hist_counts)]
    df["diff_from_median"] = np.abs(lengths - median_length)

    # Get quality scores for each sequence
    all_quality_array = []
    for i in range(len(df)):
        all_quality_array.append(_get_quality_scores(df.iloc[i,:],key_species))
    all_quality_array = np.array(all_quality_array)

    unique_species = np.unique(df.species)

    if not silent:
        P = tqdm
        print("Removing redundancy within species.",flush=True)
    else:
        P = _DummyTqdm

    with P(total=np.sum(df.keep)) as pbar:

        for i, s in enumerate(unique_species):

            species_mask = np.logical_and(df.keep,df.species == s)
            if np.sum(species_mask) < 2:
                pbar.update(np.sum(species_mask))
                continue


            # List of all sequences with keep = True as integers
            sequence_array = np.array(df.loc[species_mask,"sequence"])
            quality_array = all_quality_array[species_mask]
            keep_array = np.ones(len(sequence_array),dtype=bool)

            _reduce_redundancy_thread_manager(sequence_array=sequence_array,
                                              quality_array=quality_array,
                                              keep_array=keep_array,
                                              cutoff=cutoff,
                                              discard_key=True,
                                              num_threads=num_threads,
                                              progress_bar=False)

            # Update keep in dataframe
            df.loc[species_mask,"keep"] = keep_array

            pbar.update(np.sum(species_mask))


    if only_in_species:
        if not silent:
            final_keep_number = np.sum(df.keep)
            print(f"Reduced {starting_keep_number} --> {final_keep_number} sequences.",flush=True)
            print("Done.",flush=True)

        return df


    sequence_array = np.array(df.loc[df.keep,"sequence"])
    quality_array = all_quality_array[df.keep]
    keep_array = np.ones(len(sequence_array),dtype=bool)

    progress_bar = not silent
    if not silent:
        print("Removing redundancy in all-on-all comparison.",flush=True)

    _reduce_redundancy_thread_manager(sequence_array=sequence_array,
                                      quality_array=quality_array,
                                      keep_array=keep_array,
                                      cutoff=cutoff,
                                      discard_key=False,
                                      num_threads=num_threads,
                                      progress_bar=progress_bar)

    # Update keep array
    df.loc[df.keep,"keep"] = keep_array

    if not silent:
        final_keep_number = np.sum(df.keep)
        print(f"Reduced {starting_keep_number} --> {final_keep_number} sequences.",flush=True)
        print("Done.",flush=True)

    return df

def find_cutoff(df,
                min_cutoff=0.85,
                max_cutoff=1.00,
                try_n_values=8,
                target_number=500,
                sample_size=200,
                key_species=[]):
    """
    Find a sequence identity cutoff that yields approximately `target_number`
    sequences.

    Parameters
    ----------
    min_cutoff : float, default=0.85
        minimum identity cutoff to use (between 0 and 1)
    max_cutoff : float, default=1.00
        maximum identity cutoff to use (between 0 and 1)
    try_n_values : int, default=8
        try this many different cutoffs between min and max cutoff
    target_number : int, default=500
        find a cutoff that gets approximately this number of sequences
    sample_size : int, default=200
        grab this number of sequences from the dataframe to find the cutoff
    key_species : list, default=[]
        key species to pass to remove_redundancy

    Returns
    -------
    cutoff : float
        redundancy cutoff that yields approximately target_number sequences
    """

    df = check.check_topiary_dataframe(df)
    min_cutoff = check.check_float(min_cutoff,
                                               "min_cutoff",
                                               minimum_allowed=0,
                                               maximum_allowed=1)
    max_cutoff = check.check_float(max_cutoff,
                                               "max_cutoff",
                                               minimum_allowed=0,
                                               maximum_allowed=1)
    try_n_values = check.check_int(try_n_values,
                                               "try_n_values",
                                               minimum_allowed=2)
    target_number = check.check_int(target_number,
                                                "target_number",
                                                minimum_allowed=1)
    sample_size = check.check_int(sample_size,
                                              "sample_size",
                                              minimum_allowed=1)
    key_species = check.check_iter(key_species,
                                               "key_species",
                                               required_value_type=str,
                                               is_not_type=[str,dict])

    # Deal with cutoffs
    if min_cutoff > max_cutoff:
        err = "\nmin_cutoff must be less than or equal to max_cutoff\n\n"
        raise ValueError(err)

    # If min_cutoff and max_cutoff are the same, just return it
    if min_cutoff == max_cutoff:
        return min_cutoff
    else:
        step = (max_cutoff - min_cutoff)/(try_n_values - 1)
        cutoffs = np.array([min_cutoff + step*i for i in range(try_n_values)])

    # Make sure sample size is sane
    if sample_size > len(df.index):
        sample_size = len(df.index)

    # Create reduced dataframe to play with
    to_take = np.random.choice(df.index,sample_size,replace=False)
    small_df = df.loc[to_take,:]

    predicted_keep = []
    with tqdm(total=len(cutoffs)) as pbar:
        for c in cutoffs:
            lower_df = remove_redundancy(small_df,
                                         cutoff=c,
                                         key_species=key_species,
                                         silent=True)
            fx_kept = np.sum(lower_df.keep)/np.sum(small_df.keep)
            predicted_keep.append(fx_kept*np.sum(df.keep))

            pbar.update(1)

    predicted_keep = np.array(predicted_keep)

    # If target_number is smaller than what we got with min cutoff, return min
    # cutoff. That's as low as we can go.
    if predicted_keep[0] >= target_number:
        return min_cutoff

    # If our target number is bigger than our predicted number with the max
    # cutoff, return the maxium cutoff
    if predicted_keep[-1] <= target_number:
        return max_cutoff

    # Get the predicted_keep indexe just less than the target
    try:
        bottom = np.where(predicted_keep < target_number)[0][-1]
    except IndexError:
        bottom = None

    # Get the predicted_keep index just more than the target
    try:
        top = np.where(predicted_keep >= target_number)[0][0]
    except IndexError:
        top = None

    # Note: top and bottom should never be None at the same time because we
    # first validated that we were not entirely above the target number above.

    # If we failed to find bottom, just return top
    if bottom is None:
        return cutoffs[top]

    # If we failed to find top, just return bottom
    if top is None:
        return cutoffs[bottom]

    # Interpolate between the values just above and below the top to get cutoff
    rise = (cutoffs[top] - cutoffs[bottom])
    run = (predicted_keep[top] - predicted_keep[bottom])
    slope = rise/run
    intercept = cutoffs[top] - slope*predicted_keep[top]

    return target_number*slope + intercept



"""
Functions for cleaning up an alignment, removing poorly aligned sequences.
"""

import topiary
from topiary import check

import pandas as pd
import numpy as np

import re

# structures to convert amino acid sequences to integers and back
AA = "ACDEFGHIKLMNPQRSTVWY-"
AA_TO_INT = dict([(a,i) for i, a in enumerate(AA)])
INT_TO_AA = list(AA)

def _get_sparse_columns(seqs,sparse_column_cutoff=0.5):
    """
    Get True/False array for whether each column is less than cutoff % gaps.

    Parameters
    ----------
        seqs: 2D numpy array holding sequences represented as integers
        sparse_column_cutoff: column is sparse if it has > sparse_column_cutoff
                              fraction gap

    Return
    ------
        True/False numpy array mask
    """

    column_scores = []
    for c in range(seqs.shape[1]):
        column_scores.append(np.sum(seqs[:,c] == 20)/seqs.shape[0])
    column_scores = np.array(column_scores)

    sparse_columns = column_scores >= sparse_column_cutoff

    return sparse_columns

def _rle(input_array):
    """
    Get run-length encoding of an array. For input_array, extract runs of
    identical values.

    Parameters
    ----------
        input_array: numpy input array
        get_only_true: if set to True, return only runs of "True" from
                       input_array.

    Return
    ------
        run lengths, run start_positions, run values

    This is a cleaned up version of a solution posted here:
    https://stackoverflow.com/questions/1066758/find-length-of-sequences-of-identical-values-in-a-numpy-array-run-length-encodi
    """

    N = input_array.shape[0]

    # Get differences in input array
    diffs = input_array[1:] != input_array[:-1]

    # Build array to hold indexes of differences. Stick -1 at front and
    # len(input_array) -1 at back so we count from start and end.
    diff_indexes = np.zeros(np.sum(diffs) + 2,dtype=int)
    diff_indexes[0] = -1
    diff_indexes[-1] = N - 1

    # Get the indexes where the sequences differ from each other.
    diff_indexes[1:-1] = np.where(diffs)[0]

    # Get lengths of jumps between indexes
    run_lengths = np.diff(diff_indexes)

    # Start positions array. Append 0 on front -- first always starts at zero.
    # Don't include last index -- not a start.
    start_positions = np.zeros(run_lengths.shape[0],dtype=int)
    start_positions[1:] = np.cumsum(run_lengths[:-1])

    # These are the values of each of the runs
    values = input_array[diff_indexes[1:]]

    return run_lengths, start_positions, values

def _drop_gaps_only(seqs):
    """
    Drop gap-only columns from an array holding a sequence alignment.

    Parameters
    ----------
        seqs: 2D numpy array holding sequences represented as integers

    Return
    ------
        Copy of seqs array with gaps-only columns removed.
    """

    # Create True/False mask for columns with more than just "-"
    not_just_gaps = []
    for i in range(seqs.shape[1]):
        u = np.unique(seqs[:,i])
        if len(u) == 1 and u[0] == 20:
            not_just_gaps.append(False)
        else:
            not_just_gaps.append(True)

    not_just_gaps = np.array(not_just_gaps,dtype=bool)

    # Whack out columns that are only "-"
    seqs = seqs[:,not_just_gaps]

    return seqs

def _find_too_many_sparse(seqs,
                          cumulative_keep,
                          force_keep,
                          sparse_column_cutoff,
                          maximum_sparse_allowed):
    """
    Filter out sequences where over maximum_sparse_allowed of sequence is in
    sparse columns.

    Parameters
    ----------
        seqs: array of sequences
        cumulative_keep: array of indexes to be passed on for further analyses
        force_keep: array of bool forcing specific columns to be kept
        sparse_column_cutoff: a column is called sparse if it has
                              > sparse_column_cutoff gaps
        maximum_sparse_allowed: only keep sequences where less than
                                maximum_sparse_allowed of the sequence (fraction)
                                is a sparse column. Between 0 and 1.

    Return
    ------
        kept sequences, updated cumulative_keep, updated force_keep
    """

    sparse_columns = _get_sparse_columns(seqs,sparse_column_cutoff)
    non_gap_sparse = np.logical_and(seqs != 20,sparse_columns)
    seq_score = np.sum(non_gap_sparse,axis=1)/seqs.shape[1]

    keep_mask = seq_score <= maximum_sparse_allowed
    keep_mask = np.logical_or(keep_mask,force_keep)
    force_keep = force_keep[keep_mask]
    cumulative_keep = cumulative_keep[keep_mask]
    seqs = seqs[keep_mask]
    seqs = _drop_gaps_only(seqs)

    return seqs, cumulative_keep, force_keep

def _find_too_few_dense(seqs,
                        cumulative_keep,
                        force_keep,
                        sparse_column_cutoff,
                        minimum_dense_required):
    """
    Filter out sequences that have a non gap in less than minimum_dense_required
    fraction of the dense columns.

    Parameters
    ----------
        seqs: array of sequences
        cumulative_keep: array of indexes to be passed on for further analyses
        force_keep: array of bool forcing specific columns to be kept
        sparse_column_cutoff: a column is called sparse if it has
                              > sparse_column_cutoff gaps
        minimum_dense_required: only keep sequences that have sequence covering
                                >= than minimum_dense_required of the
                                dense columns. Between 0 and 1.

    Return
    ------
        kept sequences, updated cumulative_keep, updated force_keep
    """

    sparse_columns = _get_sparse_columns(seqs,sparse_column_cutoff)
    dense_columns = np.logical_not(sparse_columns)
    gap_dense = np.logical_and(seqs != 20,dense_columns)
    seq_score = np.sum(gap_dense,axis=1)/np.sum(dense_columns)

    keep_mask = seq_score >= minimum_dense_required
    keep_mask = np.logical_or(keep_mask,force_keep)
    force_keep = force_keep[keep_mask]
    cumulative_keep = cumulative_keep[keep_mask]
    seqs = seqs[keep_mask]
    seqs = _drop_gaps_only(seqs)

    return seqs, cumulative_keep, force_keep

def _find_long_insertions(seqs,
                          cumulative_keep,
                          force_keep,
                          sparse_column_cutoff,
                          long_insertion_length,
                          long_insertion_fx_cutoff):
    """
    Filter out sequences that are part of long stretches of sparse sequences.
    Any sequence with long_insertion_fx_cutoff of the sparse columns not
    gapped will be removed.

    Parameters
    ----------
        seqs: array of sequences
        cumulative_keep: array of indexes to be passed on for further analyses
        force_keep: array of bool forcing specific columns to be kept
        sparse_column_cutoff: a column is called sparse if it has
                              > sparse_column_cutoff gaps
        long_insertion_length: look at sequences that participate in an insertion
                               with >= than long_insertion_length contiguous
                               sparse columns.
        long_insertion_fx_cutoff: remove sequences that have sequence covering
                                  >= than long_insertion_fx_cutof of the
                                  insertion. Between 0 and 1. Default: 0.8.

    Return
    ------
        kept sequences, updated cumulative_keep, updated force_keep
    """

    sparse_columns = _get_sparse_columns(seqs,sparse_column_cutoff)
    dense_columns = np.logical_not(sparse_columns)
    num_dense_columns = np.sum(dense_columns)

    run_lengths, start_positions, values = _rle(dense_columns)

    keep_mask = np.ones(seqs.shape[0],dtype=bool)
    for i in range(len(run_lengths)):

        # If the this is a run of sparse columns
        if values[i] == False:

            # If the run length is less than the long_insertion_length stop
            # considering it
            if run_lengths[i] < long_insertion_length:
                continue

            # Grab indexes for start and stop of sparse column run
            s = start_positions[i]
            e = s + run_lengths[i]

            # Go through every sequence
            for j in range(len(seqs)):

                # If we haven't already tossed this sequence...
                if keep_mask[j]:
                    # Calculate the fraction of non-gap characters for this
                    # sequence in the sparse region. If this is greater than
                    # long_insertion_fx_cutoff, toss the sequence
                    fx_in_sparse = np.sum(seqs[j,s:e] != 20)/run_lengths[i]
                    if fx_in_sparse >= long_insertion_fx_cutoff:
                        keep_mask[j] = False

    keep_mask = np.logical_or(keep_mask,force_keep)
    force_keep = force_keep[keep_mask]
    cumulative_keep = cumulative_keep[keep_mask]
    seqs = seqs[keep_mask]
    seqs = _drop_gaps_only(seqs)

    return seqs, cumulative_keep, force_keep

def score_alignment(df,
                    alignment_column,
                    sparse_column_cutoff=0.95,
                    align_trim=(0.1,0.9)):
    """
    Calculate alignment quality scores for each sequence in the dataframe. The
    resulting scores are loaded as columns into the dataframe. In all cases, a
    higher score is a worse alignment.

    The three calculated scores are:

    + `fx_in_sparse`: fraction of columns from the sequence that have no gap
      where most other sequences have a gap.
    + `fx_missing_dense`: fraction of columns from the sequence that have a gap
      where most other sequences have no gap.
    + `sparse_run_length`: length of longest insertion in the sequence. The
      insertion does not have to be continuous in sequence. Instead, this
      function finds runs of sparse_columns and then counts how many columns
      within each run come from this sequence. If, for example, there are ten
      sparse columns in a row and a sequence has eight non-gap characters
      over that run of sparse columns, the run length would be eight.

    Parameters
    ----------
    df : pandas.DataFrame
        topiary dataframe
    alignment_column : str, default="alignment"
        column in dataframe to look for the alignment
    sparse_column_cutoff : float, default=0.95
        when checking alignment quality, a column is sparse if it has gaps in
        more than sparse_column_cutoff sequences.
    align_trim : tuple, default=(0.05,0.95)
        when checking alignment quality, do not score the first and last parts
        of the alignment. Interpreted like a slice, but with percentages.
        (0.0,1.0) would not trim; (0.05,0,98) would trim the first 0.05 off the
        front and the last 0.02 off the back.

    Returns
    -------
    topiary_dataframe : pandas.DataFrame
        copy of df with three new columns containing scores.
    """

    # check dataframe
    df = check.check_topiary_dataframe(df)

    # Check alignment column validity
    try:
        aln = df.loc[df.keep,alignment_column]
    except KeyError:
        err = f"\ndataframe does not have alignment_column '{alignment_column}'\n\n"
        raise ValueError(err)

    # Get length of sequence column
    try:
        col_lengths = list(set([len(s) for s in aln]))
        if len(col_lengths) != 1:
            err = f"\nall sequences in '{alignment_column}' do not have the same legnth\n\n"
            raise ValueError(err)
        align_length = col_lengths[0]

    except TypeError:
        err = f"\n'{alignment_column}' could not be interpreted as sequences\n\n"
        raise ValueError(err)

    # Check float inputs
    sparse_column_cutoff = check.check_float(sparse_column_cutoff,
                                                         "sparse_column_cutoff",
                                                         minimum_allowed=0,
                                                         maximum_allowed=1)

    align_trim = check.check_iter(align_trim,"align_trim",
                                              minimum_allowed=2,
                                              maximum_allowed=2)

    front_trim = check.check_float(align_trim[0],
                                               "align_trim[0]",
                                               minimum_allowed=0.0,
                                               maximum_allowed=1.0)

    back_trim =  check.check_float(align_trim[1],
                                               "align_trim[1]",
                                               minimum_allowed=0.0,
                                               maximum_allowed=1.0)

    if front_trim >= back_trim:
        err = "\nfront_trim must not overlap with back_trim\n\n"
        raise ValueError(err)

    # Get indexes for front and back trimming
    front_index = int(round(align_length*front_trim,0))
    back_index = int(round(align_length*back_trim,0))
    if front_index == back_index:
        if front_index > 0:
            front_index -= 1
        else:
            back_index += 1

    # Generate an array of sequences as integers.
    seqs = []
    for i in df.index:
        if not df.loc[i,"keep"]:
            continue
        this_seq = re.sub(f"[^{AA}]","-",df.loc[i,alignment_column][front_index:back_index])
        seqs.append([AA_TO_INT[c] for c in list(this_seq)])

    seqs = np.array(seqs,dtype=int)

    # Drop gaps only columns
    seqs = _drop_gaps_only(seqs)

    sparse_columns = _get_sparse_columns(seqs,sparse_column_cutoff)
    non_gap_sparse = np.logical_and(seqs != 20,sparse_columns)
    fx_in_sparse = np.sum(non_gap_sparse,axis=1)/seqs.shape[1]

    dense_columns = np.logical_not(sparse_columns)
    gap_dense = np.logical_and(seqs == 20,dense_columns)
    fx_missing_dense = np.sum(gap_dense,axis=1)/np.sum(dense_columns)

    sparse_run_length = np.zeros(len(seqs))
    run_lengths, start_positions, values = _rle(dense_columns)
    for i in range(len(run_lengths)):

        # If the this is a run of sparse columns
        if values[i] == False:

            # Grab indexes for start and stop of sparse column run
            s = start_positions[i]
            e = s + run_lengths[i]

            # Go through every sequence
            for j in range(len(seqs)):
                in_sparse_run = np.sum(seqs[j,s:e] != 20)
                if in_sparse_run > sparse_run_length[j]:
                    sparse_run_length[j] = in_sparse_run


    # Load quality data into the dataframe
    df["fx_in_sparse"] = np.nan
    df.loc[df.keep,"fx_in_sparse"] = fx_in_sparse
    df["fx_missing_dense"] = np.nan
    df.loc[df.keep,"fx_missing_dense"] = fx_missing_dense
    df["sparse_run_length"] = np.nan
    df.loc[df.keep,"sparse_run_length"] = sparse_run_length

    return df

def clean_alignment(df,
                    alignment_column="alignment",
                    key_species=[],
                    sparse_column_cutoff=0.95,
                    maximum_sparse_allowed=0.025,
                    minimum_dense_required=0.95,
                    long_insertion_length=8,
                    long_insertion_fx_cutoff=0.6):
    """
    Remove poorly aligned sequences from a topiary dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        topiary dataframe with alignment.
    alignment_column : str, default="alignment"
        column in dataframe to look for the alignment
    key_species : list, default=[]
        list of species (binomial names) that will not have sequences removed
        regardless of alignment quality.
    sparse_column_cutoff : float, default=0.95
        when checking alignment quality, a column is sparse if it has gaps in
        more than sparse_column_cutoff sequences.
    maximum_sparse_allowed : float, default=0.025
        only keep sequences where less than maximum_sparse_allowed of the
        sequence (fraction) is a sparse column. Between 0 and 1.
    minimum_dense_required : float, default=0.95
        only keep sequences that have sequence covering >=
        than minimum_dense_required of the dense columns. Between 0 and 1.
    long_insertion_length : int, default=8
        look at sequences that participate in an insertion with >= than
        long_insertion_length contiguous sparse columns.
    long_insertion_fx_cutoff : float, default=0.6
        remove sequences that have sequence covering >= than
        long_insertion_fx_cutof of the insertion. Between 0 and 1.

    Return
    ------
    topiary_dataframe : pandas.DataFrame
        dataframe with cleaned up alignment. sequences removed from the
        alignment have keep=False and their aligned sequence set to pandas.NA.
    """

    # check dataframe
    df = check.check_topiary_dataframe(df)

    # Check alignment column validity
    try:
        df.loc[:,alignment_column]
    except KeyError:
        err = f"\ndataframe does not have alignment_column '{alignment_column}'\n\n"
        raise ValueError(err)


    key_species = check.check_iter(key_species)

    # Check numerical arguments

    sparse_column_cutoff = check.check_float(sparse_column_cutoff,
                                                         "sparse_column_cutoff",
                                                         minimum_allowed=0,
                                                         maximum_allowed=1)

    maximum_sparse_allowed = check.check_float(maximum_sparse_allowed,
                                                           "maximum_sparse_allowed",
                                                           minimum_allowed=0,
                                                           maximum_allowed=1)

    long_insertion_length = check.check_int(long_insertion_length,
                                                        "long_insertion_length",
                                                        minimum_allowed=0)

    long_insertion_fx_cutoff = check.check_float(long_insertion_fx_cutoff,
                                                             "long_insertion_fx_cutoff",
                                                             minimum_allowed=0,
                                                             maximum_allowed=1)

    # Convert alignment into array of integers, only considering sequences we
    # were keeping already
    seqs = []
    force_keep = []

    # If the "always_keep" columns exists, keep it
    try:
        always = np.array(df.loc[:,"always_keep"],dtype=bool)
    except KeyError:
        always = np.zeros(len(df),dtype=bool)

    termini = int(round(len(df.loc[df.index[0],alignment_column])*0.05,0))

    for i in df.index:
        if not df.loc[i,"keep"]:
            continue
        this_seq = re.sub(f"[^{AA}]","-",df.loc[i,alignment_column])
        seqs.append([AA_TO_INT[c] for c in list(this_seq)])

        # If the sequence is from a key species, set force_keep to True for that
        # sequence
        if df.loc[i,"species"] in key_species or always[i]:
            force_keep.append(True)
        else:
            force_keep.append(False)



    seqs = np.array(seqs,dtype=int)
    force_keep = np.array(force_keep,dtype=bool)

    starting_keep_number = np.sum(df.keep)

    # Will hold indexes to keep after filtering, getting rid of those with
    # keep = False from the start
    cumulative_keep = df.index[df.keep]

    # Remove sequences with long runs of non-gap characters in low-quality
    # regions. (Basically, sequences with long, unique insertions)
    seqs, cumulative_keep, force_keep = _find_too_many_sparse(seqs,
                                                              cumulative_keep,
                                                              force_keep,
                                                              sparse_column_cutoff,
                                                              maximum_sparse_allowed)

    # Remove sequences with that do not cover enough of the dense columns
    # (basically, partial sequences)
    seqs, cumulative_keep, force_keep = _find_too_few_dense(seqs,
                                                            cumulative_keep,
                                                            force_keep,
                                                            sparse_column_cutoff,
                                                            minimum_dense_required)


    # Remove sequences that have long stretches of sequence in runs of sparse
    # columns
    seqs, cumulative_keep, force_keep = _find_long_insertions(seqs,
                                                              cumulative_keep,
                                                              force_keep,
                                                              sparse_column_cutoff,
                                                              long_insertion_length,
                                                              long_insertion_fx_cutoff)

    # Construct final set of sequences as strings
    final_seqs = []
    for i in range(seqs.shape[0]):
        final_seqs.append("".join([INT_TO_AA[s] for s in seqs[i,:]]))

    # Construct final dataframe with cleaned up alignment
    df.loc[:,"keep"] = False
    df.loc[cumulative_keep,"keep"] = True
    #df.loc[cumulative_keep,alignment_column] = final_seqs
    no_keep = np.array(list(set(df.index).difference(set(cumulative_keep))))
    df.loc[no_keep,alignment_column] = pd.NA

    final_keep_number = np.sum(df.keep)

    print("Removing excessively gapped sequences.")
    print(f"Reduced {starting_keep_number} --> {final_keep_number} sequences.",flush=True)

    return df



"""
Interface to muscle, compatible with muscle 3.8 and 5.1.
"""

import topiary
from topiary import check
import pandas as pd
import subprocess, sys, os, random, string, re

def run_muscle(input,
               output_fasta=None,
               super5=False,
               muscle_cmd_args=[],
               muscle_binary="muscle"):
    """
    Run muscle to align sequences.

    Parameters
    ----------
    input : str or pandas.DataFrame
        input to align (fasta file or topiary df)
    output_fasta : str or None, default=None
        output fasta file to store alignment. Optional if the input
        is a dataframe; reqiured if the input is a fasta file.
    super5 : bool, default=False
        User the 'super5' mode of muscle 5. If using muscle 3.x, this parameter
        is ignored.
    muscle_cmd_args : list
        arguments to pass directly to muscle. Wrapper specifies "-align" and
        "-output" (or -in/-out for 3.x command line), but leaves rest as
        default. Format should be something like ['-replicates',20,...].
        Arguments are not checked by this function, but passed directly to
        muscle.
    muscle_binary : str, default="muscle"
        location of muscle binary (default assumes 'muscle' command is in $PATH).

    Returns
    -------
    alignment : None or pandas.DataFrame
        If input is a topiary dataframe, return a copy of the dataframe with the
        aligned sequences loaded into the `alignment` column. Otherwise, write
        to `output_fasta file` and return None from the function.
    """

    # If output_fasta is defined, make sure it's a string
    if output_fasta and type(output_fasta) is not str:
        err = "\noutput_fasta '{output_fasta}' should be a string indicating\n"
        err += "fasta file where alignment should be written out.\n"
        raise ValueError(err)

    # If the input is a string.
    if type(input) is str:

        # Make sure the input file exists.
        if not os.path.exists(input):
            err = f"\n'{input}' file does not exist.\n\n"
            raise FileNotFoundError(err)

        # Make sure the output file is not None.
        if output_fasta is None:
            err = "\nIf aligning a fasta file, an output_fasta file must be\n"
            err += "specified.\n"
            raise ValueError(err)

        # Do the alignment.
        _run_muscle(input,output_fasta,super5,muscle_cmd_args,muscle_binary)

        print(f"\nSuccess. Alignment written to '{output_fasta}'.",flush=True)

        return None

    # If the input is a dataframe
    elif type(input) is pd.DataFrame:

        # Check the dataframe
        df = check.check_topiary_dataframe(input)

        # Create temporary input file
        tmp_file_root = "".join([random.choice(string.ascii_letters) for i in range(10)])
        input_fasta = "topiary-tmp_{}_align-in.fasta".format(tmp_file_root)
        topiary.write_fasta(df,input_fasta)

        # Create temporary output file name if required.
        temporary_output = False
        if output_fasta is None:
            output_fasta = "topiary-tmp_{}_align-out.fasta".format(tmp_file_root)
            temporary_output = True

        # Do the alignment
        _run_muscle(input_fasta,output_fasta,super5,muscle_cmd_args,muscle_binary)

        # Read alignment back into the dataframe
        df = topiary.read_fasta_into(df,output_fasta)

        # Delete temporary input file
        try:
            os.remove(input_fasta)
        except FileNotFoundError:
            pass

        # Delete temporary output file
        print("\nSuccess. Alignment written to the `alignment` column in the dataframe.",flush=True)
        if temporary_output:
            try:
                os.remove(output_fasta)
            except FileNotFoundError:
                pass
        else:
            print(f"\nSuccess. Alignment written to '{output_fasta}'.",flush=True)

        return df

    else:
        err = "\ninput should either be a fasta file or a topiary dataframe\n\n"
        raise ValueError(err)


def _run_muscle(input_fasta,
                output_fasta,
                super5=False,
                muscle_cmd_args=[],
                muscle_binary="muscle"):
    """
    Run muscle for sequence alignment.

    Parameters
    ----------
        input_fasta: input fasta file to align
        output_fasta: output fasta file to store alignment
        super5: bool. User the 'super5' mode of muscle 5
        muscle_cmd_args: list of arguments to pass directly to muscle. Wrapper
                         specifies "-align" and "-output" (or -in/-out for old
                         version of the command line), but leaves rest as default.
                         Format should be something like ['-replicates',20,...].
                         Arguments are not checked by this function, but passed
                         directly to muscle.
        muscle_binary: location of muscle binary (default assumes 'muscle'
                       command is in the PATH).

    Return
    ------
        None
    """


    # Make sure that muscle is in the path
    try:
        subprocess.run([muscle_binary])
    except FileNotFoundError:
        err = f"\nmuscle binary '{muscle_binary}' not found in path\n\n"
        raise ValueError(err)

    # This checks for the version of the muscle command line in use. The old
    # version used -version and will throw an error with this call; the current
    # version uses --version and this will work. This is not the most robust
    # check of versions, but is likely more robust than some kind of regex given
    # variety of compiled versions out there...
    ret_version = subprocess.run([muscle_binary,"--version"])
    if ret_version.returncode == 0:
        cmd = [muscle_binary,"-align",input_fasta,"-output",output_fasta]
        if super5:
            cmd[1] = "-super5"
    else:
        cmd = [muscle_binary,"-in",input_fasta,"-out",output_fasta]


    cmd.extend(muscle_cmd_args)

    # This relatively complicated code is effectively subprocess.run, but
    # captures and dumps output to stdout as it is generated. This is useful
    # for a jupyter notebook, as the output appears in the notebook rather than
    # on the command line.
    popen = subprocess.Popen(cmd,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             universal_newlines=True)
    for line in popen.stdout:
        if re.search("100.0%",line):
            endl = "\n"
        else:
            print(90*" ",end="\r")
            endl = "\r"
        print(line.strip(),end=endl,flush=True)
    popen.stdout.close()
    return_code = popen.wait()

    # If running muscle failed...
    if return_code != 0:
        raise subprocess.CalledProcessError(return_code, cmd)



"""
Functions to check/process arguments in topiary functions.
"""
from .paralog_patterns import check_paralog_patterns
from .topiary_dataframe import check_topiary_dataframe
from .standard import check_bool, check_float, check_int, check_iter



"""
Shared functions for running pipelines.
"""

import topiary

def _run_and_print(function,kwargs,step_counter,out_file_string,human_string):
    """
    Wrapper function that calls another function in a stereotyped way.

    Parameters
    ----------
        function: function to call. code assumes this function returns a
                  topiary dataframe.
        kwargs: keyword arguments to pass to the function
        step_counter: number output file with this counter
        out_file_string: write the resulting dataframe to a csv file with
                        the format step_counter_out_file_string-datframe.csv
        human_string: Write this to stdout for human readability.

    Return
    ------
        dataframe output of function, incremented step counter
    """

    print("-------------------------------------------------------------------")
    print(human_string)
    print("-------------------------------------------------------------------")
    print("",flush=True)

    df = function(**kwargs)
    topiary.write_dataframe(df,f"{step_counter:02d}_{out_file_string}-dataframe.csv")

    print("",flush=True)

    return df, step_counter + 1



"""
Functions for doing quality control on sequences in a dataframe.
"""

from .redundancy import remove_redundancy, find_cutoff
from .alignment import clean_alignment, score_alignment
from .taxonomic import taxonomic_sample



"""
Functions to check/process `bool`, `float`, `int`, and iterable arguments in
topiary functions.
"""

import numpy as np

def check_bool(value,variable_name=None):
    """
    Process a `bool` argument and do error checking.

    Parameters
    ----------
    value : 
        input value to check/process
    variable_name : str
        name of variable (string, for error message)

    Returns
    -------
    bool
        validated/coerced bool

    Raises
    ------
    ValueError
        If value cannot be interpreted as a bool
    """

    try:

        # See if this is an iterable
        if hasattr(value,"__iter__"):
            raise ValueError

        # See if this is a naked type
        if type(value) is type:
            raise ValueError

        if value != 0:
            if not np.isclose(round(value,0)/value,1):
                raise ValueError

        value = bool(int(value))

    except (TypeError,ValueError):

        if variable_name is not None:
            err = f"\n{variable_name} '{value}' must be True or False.\n\n"
        else:
            err = f"\n'{value}' must be True or False.\n\n"
        err += "\n\n"

        raise ValueError(value)

    return value


def check_float(value,
                  variable_name=None,
                  minimum_allowed=-np.inf,
                  maximum_allowed=np.inf,
                  minimum_inclusive=True,
                  maximum_inclusive=True):
    """
    Process a `float` argument and do error checking.

    Parameters
    ----------
    value :
        input value to check/process
    variable_name : str
        name of variable (string, for error message)
    minimum_allowed : float, default=-np.inf
        minimum allowable value for the variable
    maximum_allowed : float, default=np.inf
        maximum allowable value for the variable
    minimum_inclusive : bool, default=True
        whether lower bound is inclusive
    maximum_inclusive : bool, default=True
        whether upper bound is inclusive

    Returns
    -------
    float
        validated/coerced float

    Raises
    ------
    ValueError
        If value cannot be interpreted as a float
    """

    try:

        # Try to cast as string to an integer
        if type(value) is str:
            value = float(value)

        # See if this is an iterable
        if hasattr(value,"__iter__"):
            raise ValueError

        # See if this is a naked type
        if type(value) is type:
            raise ValueError

        value = float(value)

        if np.isnan(value):
            raise ValueError

        if minimum_inclusive:
            if value < minimum_allowed:
                raise ValueError
        else:
            if value <= minimum_allowed:
                raise ValueError

        if maximum_inclusive:
            if value > maximum_allowed:
                raise ValueError
        else:
            if value >= maximum_allowed:
                raise ValueError

    except (ValueError,TypeError):

        if minimum_inclusive:
            min_o = "<="
        else:
            min_o = "<"

        if maximum_inclusive:
            max_o = "<="
        else:
            max_o = "<"

        bounds = f"{minimum_allowed} {min_o} {variable_name} {max_o} {maximum_allowed}"

        if variable_name is not None:
            err = f"\n{variable_name} '{value}' must be a float:\n\n"
        else:
            err = f"\n'{value}' must be a float:\n\n"

        if not (minimum_allowed is None and maximum_allowed is None):
            err += bounds

        err += "\n\n"

        raise ValueError(err)

    return value

def check_int(value,
                variable_name=None,
                minimum_allowed=None,
                maximum_allowed=None,
                minimum_inclusive=True,
                maximum_inclusive=True):
    """
    Process an `int` argument and do error checking.

    Parameters
    ----------
    value :
        input value to check/process
    variable_name : str
        name of variable (string, for error message)
    minimum_allowed : float, optional
        minimum allowable value for the variable
    maximum_allowed : float, optional
        maximum allowable value for the variable
    minimum_inclusive : bool, default=True
        whether lower bound is inclusive
    maximum_inclusive : bool, default=True
        whether upper bound is inclusive

    Returns
    -------
    int
        validated/coerced integer

    Raises
    ------
    ValueError
        If value cannot be interpreted as a int
    """


    try:

        # Try to cast as string to an integer
        if type(value) is str:
            value = int(value)

        # See if this is an iterable
        if hasattr(value,"__iter__"):
            raise ValueError

        # See if this is a naked type
        if type(value) is type:
            raise ValueError

        # If this is a float to int cast, make sure it does not have decimal
        floored = np.floor(value)
        if value != floored:
            raise ValueError

        # Make int cast
        value = int(value)

        if minimum_allowed is not None:
            if minimum_inclusive:
                if value < minimum_allowed:
                    raise ValueError
            else:
                if value <= minimum_allowed:
                    raise ValueError

        if maximum_allowed is not None:
            if maximum_inclusive:
                if value > maximum_allowed:
                    raise ValueError
            else:
                if value >= maximum_allowed:
                    raise ValueError

    except (ValueError,TypeError,OverflowError):

        if minimum_inclusive:
            min_o = "<="
        else:
            min_o = "<"

        if maximum_inclusive:
            max_o = "<="
        else:
            max_o = "<"

        bounds = f"{minimum_allowed} {min_o} {variable_name} {max_o} {maximum_allowed}"

        if variable_name is not None:
            err = f"\n{variable_name} '{value}' must be an integer:\n\n"
        else:
            err = f"\n'{value}' must be an integer:\n\n"

        if not (minimum_allowed is None and maximum_allowed is None):
            err += bounds

        err += "\n\n"

        raise ValueError(err)

    return value

def check_iter(value,
                 variable_name=None,
                 required_iter_type=None,
                 required_value_type=None,
                 minimum_allowed=None,
                 maximum_allowed=None,
                 minimum_inclusive=True,
                 maximum_inclusive=True,
                 is_not_type=None):
    """
    Process an iterable argument and do error checking.

    Parameters
    ----------
    value :
        input value to check/process
    variable_name : str
        name of variable (string, for error message)
    required_iter_type : type, optional, default=None
        If not None, validate that the iterable has the specified type
    required_value_type : type, optional, default=None
        if not None, validate that every value in the iterable has the
        specified type
    minimum_allowed : int, optional, default=None
        minimum allowable length for the iterable
    maximum_allowed : int, optional, default=None
        maximum allowable length for the iterable
    minimum_inclusive : bool, default=True
        whether lower bound is inclusive
    maximum_inclusive : bool, default=True
        whether upper bound is inclusive
    is_not_type : type, list, default=None
        type (or list of types) to reject

    Returns
    -------
    iterable
        validated/coerced iterable

    Notes
    -----
    This function does *not* check required_value_type for numpy arrays.

    Raises
    ------
    ValueError
        If value cannot be interpreted as an iterable of appropriate type
    """

    # Set up error message
    if variable_name is not None:
        err_base = f"\nvariable_name '{value}' "
    else:
        err_base = f"\n'{value}' "

    # Make sure it's an iterable
    if not hasattr(value,"__iter__"):
        err = err_base + "must be list-like\n"
        raise ValueError(err)

    # See if this is a naked type
    if type(value) is type:
        err = err_base + "must not be a type\n"
        raise ValueError(err)

    # If requested, make sure it's the required type
    if required_iter_type is not None:
        if type(value) is not required_iter_type:
            err = err_base + f"must be type {required_iter_type}\n"
            raise ValueError(err)

    # If reuqested, make sure the values are of the right type
    if required_value_type is not None:

        # Skip check for iterable with no values
        if len(value) > 0:

            # Only do check iterables that are not numpy arrays
            if type(value) is not np.ndarray:
                types = list(set([type(v) for v in value]))
                if len(types) != 1 or types[0] is not required_value_type:
                    err = err_base + f"all entries must have type {required_value_type}\n"
                    raise ValueError(err)

    # If requested, make sure the iterable has appropriate dimensions
    try:
        if minimum_allowed is not None:
            if minimum_inclusive:
                if len(value) < minimum_allowed:
                    raise ValueError
            else:
                if len(value) <= minimum_allowed:
                    raise ValueError

        if maximum_allowed is not None:
            if maximum_inclusive:
                if len(value) > maximum_allowed:
                    raise ValueError
            else:
                if len(value) >= maximum_allowed:
                    raise ValueError
    except ValueError:

        if minimum_inclusive:
            min_o = "<="
        else:
            min_o = "<"

        if maximum_inclusive:
            max_o = "<="
        else:
            max_o = "<"

        bounds = f"{minimum_allowed} {min_o} {variable_name} {max_o} {maximum_allowed}"

        err = err_base + f"must have length:\n\n{bounds}\n\n"
        raise ValueError(err)

    # Make sure iterable does not have a specified excluded type
    if is_not_type is not None:
        if type(is_not_type) is type:
            is_not_type = [is_not_type]

        if type(value) in is_not_type:
            bad_types = ",".join([f"{v}" for v in is_not_type])

            err = err_base + f"must not be of type {bad_types}\n\n"
            raise ValueError(err)

    return value



"""
Draw a species tree with tips labeled by species name.
"""

import topiary
from ._base import setup_generic_tree_format, final_render

import ete3


def species_tree(species_tree,
                 output_file=None,
                 fontsize=36,
                 circle_radius=0.0,
                 width=800,space_between_taxa=10,line_width=4):
    """
    Draw a species tree with tips labeled by species name.

    Parameters
    ----------
    species_tree: ete3.Tree
        ete3 tree object holding a species tree
    output_file : str, optional
        output file to write tree. If running in a notebook, will return to
        notebook and write to this file. If running in a notebook but not
        specified, do not write to a file. If not in a notebook and not
        specified, will write out ancestor-tree.pdf.
    fontsize : float, default=36
        fontsize in points for labels
    circle_radius : float, default=0.025
        circle size for internal nodes (fraction of total width)
    width : int, default=800
        width of total tree in pixels
    space_between_taxa : int, default=10
        number of pixels between taxa, sets height.
    line_width : int, default=4
        width of lines used to draw tree (pixels)
    df : pandas.DataFrame or None, default=None
        topiary dataframe (overides whatever is in run_dir/output/dataframe.csv)

    Returns
    -------
    Python.core.display.Image or None
        if running in jupyter notebook, return Image; otherwise, return None
    """

    T = species_tree.copy()

    # Set up formats (color map for main nodes, tree format (ts) and generic
    # node format (ns)).
    cm, ts, ns = setup_generic_tree_format(num_leaves=len(T.get_leaves()),
                                           width=width,
                                           space_between_taxa=space_between_taxa,
                                           line_width=line_width)

    for n in T.traverse():

        n.set_style(ns)

        if n.is_leaf():

            # Add text for clean name
            n.name = ""
            txt = ete3.TextFace(n.species[0],fsize=fontsize)
            txt.margin_left = 4
            n.add_face(txt,0,position="branch-right")


    return final_render(T,ts,output_file,"species-tree.pdf")
